import pandas as pd

class MagicWandIntervention:

    configuration_defaults = {
        'intervention': {
            'effect_size': 0.5,
        }
    }

    def __init__(self, name, affected_value):
        self.name = name
        self.affected_value = affected_value
        self.configuration_defaults = {name: MagicWandIntervention.configuration_defaults['intervention']}

    def setup(self, builder):
        effect_size = builder.configuration[self.name].effect_size
        builder.value.register_value_modifier(self.affected_value, modifier=self.intervention_effect)
        self.effect_size = builder.value.register_value_producer(
            f'{self.name}.effect_size', source=lambda index: pd.Series(effect_size, index=index)
        )

    def intervention_effect(self, index, value):
        return value * (1 - self.effect_size(index))



import pytest

from vivarium.framework.engine import SimulationContext, Builder
from vivarium.framework.artifact import ArtifactManager, ArtifactInterface
from vivarium.framework.event import EventManager, EventInterface
from vivarium.framework.lookup import LookupTableManager, LookupTableInterface
from vivarium.framework.components import OrderedComponentSet, ComponentManager, ComponentInterface
from vivarium.framework.metrics import Metrics
from vivarium.framework.population import PopulationManager, PopulationInterface
from vivarium.framework.randomness import RandomnessManager, RandomnessInterface
from vivarium.framework.resource import ResourceManager, ResourceInterface
from vivarium.framework.time import DateTimeClock, TimeInterface
from vivarium.framework.values import ValuesManager, ValuesInterface

from .components.mocks import MockComponentA, MockComponentB, Listener


def is_same_object_method(m1, m2):
    return m1.__func__ is m2.__func__ and m1.__self__ is m2.__self__


@pytest.fixture
def components():
    return [MockComponentA('gretchen', 'whimsy'),
            Listener('listener'),
            MockComponentB('spoon', 'antelope', 23)]


@pytest.fixture
def log(mocker):
    return mocker.patch('vivarium.framework.engine.logger')


def test_SimulationContext_init_default(components):
    sim = SimulationContext(components=components)

    assert isinstance(sim._component_manager, ComponentManager)
    assert isinstance(sim._clock, DateTimeClock)
    assert isinstance(sim._data, ArtifactManager)
    assert isinstance(sim._values, ValuesManager)
    assert isinstance(sim._events, EventManager)
    assert isinstance(sim._population, PopulationManager)
    assert isinstance(sim._tables, LookupTableManager)
    assert isinstance(sim._randomness, RandomnessManager)

    assert isinstance(sim._builder, Builder)
    assert sim._builder.configuration is sim.configuration
    assert isinstance(sim._builder.lookup, LookupTableInterface)
    assert sim._builder.lookup._manager is sim._tables
    assert isinstance(sim._builder.value, ValuesInterface)
    assert sim._builder.value._manager is sim._values
    assert isinstance(sim._builder.event, EventInterface)
    assert sim._builder.event._manager is sim._events
    assert isinstance(sim._builder.population, PopulationInterface)
    assert sim._builder.population._manager is sim._population
    assert isinstance(sim._builder.randomness, RandomnessInterface)
    assert sim._builder.randomness._manager is sim._randomness
    assert isinstance(sim._builder.resources, ResourceInterface)
    assert sim._builder.resources._manager is sim._resource
    assert isinstance(sim._builder.time, TimeInterface)
    assert sim._builder.time._manager is sim._clock
    assert isinstance(sim._builder.components, ComponentInterface)
    assert sim._builder.components._manager is sim._component_manager
    assert isinstance(sim._builder.data, ArtifactInterface)
    assert sim._builder.data._manager is sim._data

    # Ordering matters.
    managers = [sim._clock, sim._lifecycle, sim._resource, sim._values,
                sim._population, sim._randomness, sim._events,
                sim._tables, sim._data]
    assert sim._component_manager._managers == OrderedComponentSet(*managers)
    unpacked_components = []
    for c in components:
        unpacked_components.append(c)
        if hasattr(c, 'sub_components'):
            unpacked_components.extend(c.sub_components)
    assert list(sim._component_manager._components)[:-1] == unpacked_components
    assert isinstance(list(sim._component_manager._components)[-1], Metrics)


def test_SimulationContext_setup_default(base_config, components):
    sim = SimulationContext(base_config, components)
    listener = [c for c in components if 'listener' in c.args][0]
    assert not listener.post_setup_called
    sim.setup()

    unpacked_components = []
    for c in components:
        unpacked_components.append(c)
        if hasattr(c, 'sub_components'):
            unpacked_components.extend(c.sub_components)
    unpacked_components.append(Metrics())

    for a, b in zip(sim._component_manager._components, unpacked_components):
        assert type(a) == type(b)
        if hasattr(a, 'args'):
            assert a.args == b.args

    assert is_same_object_method(sim.simulant_creator, sim._population._create_simulants)
    assert sim.time_step_events == ['time_step__prepare', 'time_step', 'time_step__cleanup', 'collect_metrics']
    for k in sim.time_step_emitters.keys():
        assert is_same_object_method(sim.time_step_emitters[k], sim._events._event_types[k].emit)

    assert is_same_object_method(sim.end_emitter, sim._events._event_types['simulation_end'].emit)

    assert listener.post_setup_called


def test_SimulationContext_initialize_simulants(base_config, components):
    sim = SimulationContext(base_config, components)
    sim.setup()
    pop_size = sim.configuration.population.population_size
    current_time = sim._clock.time
    assert sim._population.get_population(True).empty
    sim.initialize_simulants()
    assert len(sim._population.get_population(True)) == pop_size
    assert sim._clock.time == current_time


def test_SimulationContext_step(log, base_config, components):
    sim = SimulationContext(base_config, components)
    sim.setup()
    sim.initialize_simulants()

    current_time = sim._clock.time
    step_size = sim._clock.step_size

    listener = [c for c in components if 'listener' in c.args][0]

    assert not listener.time_step__prepare_called
    assert not listener.time_step_called
    assert not listener.time_step__cleanup_called
    assert not listener.collect_metrics_called

    sim.step()

    assert log.debug.called_once_with(current_time)
    assert listener.time_step__prepare_called
    assert listener.time_step_called
    assert listener.time_step__cleanup_called
    assert listener.collect_metrics_called

    assert sim._clock.time == current_time + step_size


def test_SimulationContext_finalize(base_config, components):
    sim = SimulationContext(base_config, components)
    listener = [c for c in components if 'listener' in c.args][0]
    sim.setup()
    sim.initialize_simulants()
    sim.step()
    assert not listener.simulation_end_called
    sim.finalize()
    assert listener.simulation_end_called


def test_SimulationContext_report(base_config, components):
    sim = SimulationContext(base_config, components)
    sim.setup()
    sim.initialize_simulants()
    sim.run()
    sim.finalize()
    metrics = sim.report()
    assert metrics['test'] == len([c for c in sim._component_manager._components if isinstance(c, MockComponentB)])



import textwrap

import pytest
import yaml

from vivarium.config_tree import (ConfigNode, ConfigTree,
                                  ConfigurationError, ConfigurationKeyError, DuplicatedConfigurationError)


@pytest.fixture(params=list(range(1, 5)))
def layers(request):
    return [f'layer_{i}' for i in range(1, request.param + 1)]


@pytest.fixture
def layers_and_values(layers):
    return {layer: f'test_value_{i+1}' for i, layer in enumerate(layers)}


@pytest.fixture
def empty_node(layers):
    return ConfigNode(layers, name='test_node')


@pytest.fixture
def full_node(layers_and_values):
    n = ConfigNode(list(layers_and_values.keys()), name='test_node')
    for layer, value in layers_and_values.items():
        n.update(value, layer, source=None)
    return n


@pytest.fixture
def empty_tree(layers):
    return ConfigTree(layers=layers)


def test_node_creation(empty_node):
    assert not empty_node
    assert not empty_node.accessed
    assert not empty_node.metadata
    assert not repr(empty_node)
    assert not str(empty_node)


def test_full_node_update(full_node):
    assert full_node
    assert not full_node.accessed
    assert len(full_node.metadata) == len(full_node._layers)
    assert repr(full_node)
    assert str(full_node)


def test_node_update_no_args():
    n = ConfigNode(['base'], name='test_node')
    n.update('test_value', layer=None, source=None)
    assert n._values['base'] == (None, 'test_value')

    n = ConfigNode(['layer_1', 'layer_2'], name='test_node')
    n.update('test_value', layer=None, source=None)
    assert 'layer_1' not in n._values
    assert n._values['layer_2'] == (None, 'test_value')


def test_node_update_with_args():
    n = ConfigNode(['base'], name='test_node')
    n.update('test_value', layer=None, source='test')
    assert n._values['base'] == ('test', 'test_value')

    n = ConfigNode(['base'], name='test_node')
    n.update('test_value', layer='base', source='test')
    assert n._values['base'] == ('test', 'test_value')

    n = ConfigNode(['layer_1', 'layer_2'], name='test_node')
    n.update('test_value', layer=None, source='test')
    assert 'layer_1' not in n._values
    assert n._values['layer_2'] == ('test', 'test_value')

    n = ConfigNode(['layer_1', 'layer_2'], name='test_node')
    n.update('test_value', layer='layer_1', source='test')
    assert 'layer_2' not in n._values
    assert n._values['layer_1'] == ('test', 'test_value')

    n = ConfigNode(['layer_1', 'layer_2'], name='test_node')
    n.update('test_value', layer='layer_2', source='test')
    assert 'layer_1' not in n._values
    assert n._values['layer_2'] == ('test', 'test_value')

    n = ConfigNode(['layer_1', 'layer_2'], name='test_node')
    n.update('test_value', layer='layer_1', source='test')
    n.update('test_value', layer='layer_2', source='test')
    assert n._values['layer_1'] == ('test', 'test_value')
    assert n._values['layer_2'] == ('test', 'test_value')


def test_node_frozen_update():
    n = ConfigNode(['base'], name='test_node')
    n.freeze()
    with pytest.raises(ConfigurationError):
        n.update('test_val', layer=None, source=None)


def test_node_bad_layer_update():
    n = ConfigNode(['base'], name='test_node')
    with pytest.raises(ConfigurationKeyError):
        n.update('test_value', layer='layer_1', source=None)


def test_node_duplicate_update():
    n = ConfigNode(['base'], name='test_node')
    n.update('test_value', layer=None, source=None)
    with pytest.raises(DuplicatedConfigurationError):
        n.update('test_value', layer=None, source=None)


def test_node_get_value_with_source_empty(empty_node):
    with pytest.raises(ConfigurationKeyError):
        empty_node._get_value_with_source(layer=None)

    for layer in empty_node._layers:
        with pytest.raises(ConfigurationKeyError):
            empty_node._get_value_with_source(layer=layer)

    assert not empty_node.accessed


def test_node_get_value_with_source(full_node):
    assert full_node._get_value_with_source(layer=None) == (None, f'test_value_{len(full_node._layers)}')

    for i, layer in enumerate(full_node._layers):
        assert full_node._get_value_with_source(layer=layer) == (None, f'test_value_{i+1}')

    assert not full_node.accessed


def test_node_get_value_empty(empty_node):
    with pytest.raises(ConfigurationKeyError):
        empty_node.get_value(layer=None)

    for layer in empty_node._layers:
        with pytest.raises(ConfigurationKeyError):
            empty_node.get_value(layer=layer)

    assert not empty_node.accessed


def test_node_get_value(full_node):
    assert full_node.get_value(layer=None) == f'test_value_{len(full_node._layers)}'
    assert full_node.accessed
    full_node._accessed = False

    for i, layer in enumerate(full_node._layers):
        assert full_node.get_value(layer=layer) == f'test_value_{i + 1}'
        assert full_node.accessed
        full_node._accessed = False

    assert not full_node.accessed


def test_node_repr():
    n = ConfigNode(['base'], name='test_node')
    n.update('test_value', layer='base', source='test')
    s = '''\
        base: test_value
            source: test'''
    assert repr(n) == textwrap.dedent(s)

    n = ConfigNode(['base', 'layer_1'], name='test_node')
    n.update('test_value', layer='base', source='test')
    s = '''\
        base: test_value
            source: test'''
    assert repr(n) == textwrap.dedent(s)

    n = ConfigNode(['base', 'layer_1'], name='test_node')
    n.update('test_value', layer=None, source='test')
    s = '''\
        layer_1: test_value
            source: test'''
    assert repr(n) == textwrap.dedent(s)

    n = ConfigNode(['base', 'layer_1'], name='test_node')
    n.update('test_value', layer='base', source='test')
    n.update('test_value', layer='layer_1', source='test')
    s = '''\
        layer_1: test_value
            source: test
        base: test_value
            source: test'''
    assert repr(n) == textwrap.dedent(s)


def test_node_str():
    n = ConfigNode(['base'], name='test_node')
    n.update('test_value', layer='base', source='test')
    s = 'base: test_value'
    assert str(n) == s

    n = ConfigNode(['base', 'layer_1'], name='test_node')
    n.update('test_value', layer='base', source='test')
    s = 'base: test_value'
    assert str(n) == s

    n = ConfigNode(['base', 'layer_1'], name='test_node')
    n.update('test_value', layer=None, source='test')
    s = 'layer_1: test_value'
    assert str(n) == s

    n = ConfigNode(['base', 'layer_1'], name='test_node')
    n.update('test_value', layer='base', source='test')
    n.update('test_value', layer='layer_1', source='test')
    s = 'layer_1: test_value'
    assert str(n) == s


def test_tree_creation(empty_tree):
    assert len(empty_tree) == 0
    assert not empty_tree.items()
    assert not empty_tree.values()
    assert not empty_tree.keys()
    assert not repr(empty_tree)
    assert not str(empty_tree)
    assert not empty_tree._children
    assert empty_tree.to_dict() == {}


def test_tree_coerce_dict():
    d, s = {}, 'test'
    assert ConfigTree._coerce(d, s) == (d, s)
    d, s = {'key': 'val'}, 'test'
    assert ConfigTree._coerce(d, s) == (d, s)
    d = {'key1': {'sub_key1': ['val', 'val', 'val'],
                  'sub_key2': 'val'},
         'key2': 'val'}
    s = 'test'
    assert ConfigTree._coerce(d, s) == (d, s)


def test_tree_coerce_str():
    d = ''''''
    s = 'test'
    assert ConfigTree._coerce(d, s) == (None, s)
    d = '''\
    key: val'''
    assert ConfigTree._coerce(d, s) == ({'key': 'val'}, s)
    d = '''\
    key1:
        sub_key1:
            - val
            - val
            - val
        sub_key2: val
    key2: val'''
    r = {'key1': {'sub_key1': ['val', 'val', 'val'],
                  'sub_key2': 'val'},
         'key2': 'val'}
    assert ConfigTree._coerce(d, s) == (r, s)
    d = '''\
        key1:
            sub_key1: [val, val, val]
            sub_key2: val
        key2: val'''
    r = {'key1': {'sub_key1': ['val', 'val', 'val'],
                  'sub_key2': 'val'},
         'key2': 'val'}
    assert ConfigTree._coerce(d, s) == (r, s)


def test_tree_coerce_yaml(tmpdir):
    d = '''\
     key1:
         sub_key1:
             - val
             - val
             - val
         sub_key2: [val, val]
     key2: val'''
    r = {'key1': {'sub_key1': ['val', 'val', 'val'],
                  'sub_key2': ['val', 'val']},
         'key2': 'val'}
    s = 'test'
    p = tmpdir.join('model_spec.yaml')
    with p.open('w') as f:
        f.write(d)
    assert ConfigTree._coerce(str(p), s) == (r, s)
    assert ConfigTree._coerce(str(p), None) == (r, str(p))


def test_single_layer():
    d = ConfigTree()
    d.update({'test_key': 'test_value', 'test_key2': 'test_value2'})

    assert d.test_key == 'test_value'
    assert d.test_key2 == 'test_value2'

    with pytest.raises(DuplicatedConfigurationError):
        d.test_key2 = 'test_value3'

    assert d.test_key2 == 'test_value2'
    assert d.test_key == 'test_value'


def test_dictionary_style_access():
    d = ConfigTree()
    d.update({'test_key': 'test_value', 'test_key2': 'test_value2'})

    assert d['test_key'] == 'test_value'
    assert d['test_key2'] == 'test_value2'

    with pytest.raises(DuplicatedConfigurationError):
        d['test_key2'] = 'test_value3'

    assert d['test_key2'] == 'test_value2'
    assert d['test_key'] == 'test_value'


def test_get_missing_key():
    d = ConfigTree()
    with pytest.raises(ConfigurationKeyError):
        _ = d.missing_key


def test_set_missing_key():
    d = ConfigTree()
    with pytest.raises(ConfigurationKeyError):
        d.missing_key = 'test_value'
    with pytest.raises(ConfigurationKeyError):
        d['missing_key'] = 'test_value'


def test_multiple_layer_get():
    d = ConfigTree(layers=['first', 'second', 'third'])
    d._set_with_metadata('test_key', 'test_with_source_value', 'first', source=None)
    d._set_with_metadata('test_key', 'test_value2', 'second', source=None)
    d._set_with_metadata('test_key', 'test_value3', 'third', source=None)

    d._set_with_metadata('test_key2', 'test_value4', 'first', source=None)
    d._set_with_metadata('test_key2', 'test_value5', 'second', source=None)

    d._set_with_metadata('test_key3', 'test_value6', 'first', source=None)

    assert d.test_key == 'test_value3'
    assert d.test_key2 == 'test_value5'
    assert d.test_key3 == 'test_value6'


def test_outer_layer_set():
    d = ConfigTree(layers=['inner', 'outer'])
    d._set_with_metadata('test_key', 'test_value', 'inner', source=None)
    d._set_with_metadata('test_key', 'test_value3', layer=None, source=None)
    assert d.test_key == 'test_value3'
    assert d['test_key'] == 'test_value3'

    d = ConfigTree(layers=['inner', 'outer'])
    d._set_with_metadata('test_key', 'test_value', 'inner', source=None)
    d.test_key = 'test_value3'
    assert d.test_key == 'test_value3'
    assert d['test_key'] == 'test_value3'

    d = ConfigTree(layers=['inner', 'outer'])
    d._set_with_metadata('test_key', 'test_value', 'inner', source=None)
    d['test_key'] = 'test_value3'
    assert d.test_key == 'test_value3'
    assert d['test_key'] == 'test_value3'


def test_update_dict():
    d = ConfigTree(layers=['inner', 'outer'])
    d.update({'test_key': 'test_value', 'test_key2': 'test_value2'}, layer='inner')
    d.update({'test_key': 'test_value3'}, layer='outer')

    assert d.test_key == 'test_value3'
    assert d.test_key2 == 'test_value2'


def test_update_dict_nested():
    d = ConfigTree(layers=['inner', 'outer'])
    d.update({'test_container': {'test_key': 'test_value', 'test_key2': 'test_value2'}}, layer='inner')
    with pytest.raises(DuplicatedConfigurationError):
        d.update({'test_container': {'test_key': 'test_value3'}}, layer='inner')

    assert d.test_container.test_key == 'test_value'
    assert d.test_container.test_key2 == 'test_value2'

    d.update({'test_container': {'test_key2': 'test_value4'}}, layer='outer')

    assert d.test_container.test_key2 == 'test_value4'


def test_source_metadata():
    d = ConfigTree(layers=['inner', 'outer'])
    d.update({'test_key': 'test_value'}, layer='inner', source='initial_load')
    d.update({'test_key': 'test_value2'}, layer='outer', source='update')

    assert d.metadata('test_key') == [
        {'layer': 'inner', 'source': 'initial_load', 'value': 'test_value'},
        {'layer': 'outer', 'source': 'update', 'value': 'test_value2'}]


def test_exception_on_source_for_missing_key():
    d = ConfigTree(layers=['inner', 'outer'])
    d.update({'test_key': 'test_value'}, layer='inner', source='initial_load')

    with pytest.raises(ConfigurationKeyError):
        d.metadata('missing_key')


def test_unused_keys():
    d = ConfigTree({'test_key': {'test_key2': 'test_value', 'test_key3': 'test_value2'}})

    assert d.unused_keys() == ['test_key.test_key2', 'test_key.test_key3']

    _ = d.test_key.test_key2

    assert d.unused_keys() == ['test_key.test_key3']

    _ = d.test_key.test_key3

    assert not d.unused_keys()


def test_to_dict_dict():
    test_dict = {'configuration': {'time': {'start': {'year': 2000}}}}
    config = ConfigTree(test_dict)
    assert config.to_dict() == test_dict


def test_to_dict_yaml(test_spec):
    config = ConfigTree(str(test_spec))
    with test_spec.open() as f:
        yaml_config = yaml.full_load(f)
    assert yaml_config == config.to_dict()


def test_freeze():
    config = ConfigTree(data={'configuration': {'time': {'start': {'year': 2000}}}})
    config.freeze()

    with pytest.raises(ConfigurationError):
        config.update(data={'configuration': {'time': {'end': {'year': 2001}}}})


def test_retrieval_behavior():
    layer_inner = 'inner'
    layer_middle = 'middle'
    layer_outer = 'outer'

    default_cfg_value = 'value_a'

    layer_list = [layer_inner, layer_middle, layer_outer]
    # update the ConfigTree layers in different order and verify that has no effect on
    #  the values retrieved ("outer" is retrieved when no layer is specified regardless of
    #  the initialization order
    for scenario in [layer_list, reversed(layer_list)]:
        cfg = ConfigTree(layers=layer_list)
        for layer in scenario:
            cfg.update({default_cfg_value: layer}, layer=layer)
        assert cfg.get_from_layer(default_cfg_value) == layer_outer
        assert cfg.get_from_layer(default_cfg_value, layer=layer_outer) == layer_outer
        assert cfg.get_from_layer(default_cfg_value, layer=layer_middle) == layer_middle
        assert cfg.get_from_layer(default_cfg_value, layer=layer_inner) == layer_inner


def test_repr_display():
    expected_repr = '''\
    Key1:
        override_2: value_ov_2
            source: ov2_src
        override_1: value_ov_1
            source: ov1_src
        base: value_base
            source: base_src'''
    # codifies the notion that repr() displays values from most to least overridden
    #  regardless of initialization order
    layers = ['base', 'override_1', 'override_2']
    cfg = ConfigTree(layers=layers)

    cfg.update({'Key1': 'value_ov_2'}, layer='override_2', source='ov2_src')
    cfg.update({'Key1': 'value_ov_1'}, layer='override_1', source='ov1_src')
    cfg.update({'Key1': 'value_base'}, layer='base', source='base_src')
    assert repr(cfg) == textwrap.dedent(expected_repr)

    cfg = ConfigTree(layers=layers)
    cfg.update({'Key1': 'value_base'}, layer='base', source='base_src')
    cfg.update({'Key1': 'value_ov_1'}, layer='override_1', source='ov1_src')
    cfg.update({'Key1': 'value_ov_2'}, layer='override_2', source='ov2_src')
    assert repr(cfg) == textwrap.dedent(expected_repr)



import pytest
from unittest.mock import call
from pathlib import Path

from vivarium.framework.artifact.artifact import Artifact, ArtifactException, _to_tree, _parse_draw_filters


@pytest.fixture()
def keys_mock():
    keys = ['metadata.locations', 'metadata.keyspace', 'metadata.versions',
            'cause.all_causes.cause_specific_mortality', 'cause.all_causes.restrictions',
            'cause.diarrheal_diseases.cause_specific_mortality', 'cause.diarrheal_diseases.death',
            'cause.diarrheal_diseases.etiologies', 'cause.diarrheal_diseases.excess_mortality',
            'cause.diarrheal_diseases.incidence', 'cause.diarrheal_diseases.population_attributable_fraction',
            'cause.diarrheal_diseases.prevalence', 'cause.diarrheal_diseases.remission',
            'cause.diarrheal_diseases.restrictions', 'cause.diarrheal_diseases.sequelae',
            'covariate.dtp3_coverage_proportion.estimate', 'dimensions.full_space',
            'etiology.adenovirus.population_attributable_fraction',
            'etiology.aeromonas.population_attributable_fraction',
            'etiology.amoebiasis.population_attributable_fraction', 'population.age_bins', 'population.structure',
            'population.theoretical_minimum_risk_life_expectancy', 'risk_factor.child_stunting.affected_causes',
            'risk_factor.child_stunting.distribution', 'risk_factor.child_stunting.exposure',
            'risk_factor.child_stunting.levels', 'risk_factor.child_stunting.relative_risk',
            'risk_factor.child_stunting.restrictions', 'sequela.moderate_diarrheal_diseases.disability_weight',
            'sequela.moderate_diarrheal_diseases.healthstate', 'sequela.moderate_diarrheal_diseases.incidence',
            'sequela.moderate_diarrheal_diseases.prevalence', 'no_data.key']
    return keys


@pytest.fixture()
def hdf_mock(mocker, keys_mock):
    mock = mocker.patch('vivarium.framework.artifact.artifact.hdf')

    def mock_load(_, key, __, ___):
        if key in keys_mock:
            if key == 'metadata.keyspace':
                return keys_mock
            elif key != 'no_data.key':
                return 'data'
            else:
                return None

    mock.load.side_effect = mock_load
    mock.get_keys.return_value = keys_mock
    mock.touch.side_effect = lambda _: None
    return mock


# keys in test artifact
_KEYS = ['population.age_bins',
         'population.structure',
         'population.theoretical_minimum_risk_life_expectancy',
         'cause.all_causes.restrictions']


def test_artifact_creation(hdf_mock, keys_mock):
    path = Path('path/to/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']

    a = Artifact(path)

    assert a.filter_terms is None
    assert a._cache == {}
    assert a.keys == keys_mock
    hdf_mock.load.called_once_with('metadata.keyspace')

    a = Artifact(path, filter_terms)

    assert a.path == str(path)
    assert a.filter_terms == filter_terms
    assert a._cache == {}
    assert a.keys == keys_mock
    hdf_mock.load.called_once_with('metadata.keyspace')


def test_artifact_load_missing_key(hdf_mock):
    path = Path('/place/with/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']
    key = 'not.a_real.key'

    a = Artifact(path, filter_terms)
    hdf_mock.load.called_once_with('metadata.keyspace')
    hdf_mock.load.reset_mock()
    with pytest.raises(ArtifactException) as err_info:
        a.load(key)

    assert f"{key} should be in {path}." == str(err_info.value)
    hdf_mock.load.assert_not_called()
    assert a._cache == {}


def test_artifact_load_key_has_no_data(hdf_mock):
    path = Path('/place/with/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']
    key = 'no_data.key'

    a = Artifact(path, filter_terms)

    with pytest.raises(AssertionError) as err_info:
        a.load(key)

    assert f"Data for {key} is not available. Check your model specification." == str(err_info.value)
    assert hdf_mock.load.called_once_with(path, key, filter_terms)
    assert a._cache == {}


def test_artifact_load(hdf_mock, keys_mock):
    path = Path('/place/with/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']

    a = Artifact(path, filter_terms)
    keys_without_metadata = set(keys_mock)-{'metadata.locations', 'metadata.keyspace', 'metadata.versions'}
    for key in keys_without_metadata:
        if key == 'no_data.key':
            continue

        assert key not in a._cache

        result = a.load(key)

        assert hdf_mock.load.called_once_with(path, key, filter_terms)
        assert key in a._cache
        assert a._cache[key] == 'data'
        assert result == 'data'

        hdf_mock.load.reset_mock()


def test_artifact_write_duplicate_key(hdf_mock, keys_mock):
    path = Path('/place/with/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']
    key = 'population.structure'

    art = Artifact(path, filter_terms)
    initial_keys = art.keys
    assert initial_keys == keys_mock

    with pytest.raises(ArtifactException) as err_info:
        art.write(key, "data")

    assert f'{key} already in artifact.' == str(err_info.value)
    assert key in art
    assert key in art.keys
    assert key not in art._cache
    hdf_mock.write.called_once_with(path, 'metadata.keyspace', ['metadata.keyspace'])
    hdf_mock.remove.assert_not_called()
    assert art.keys == initial_keys


def test_artifact_write_no_data(hdf_mock):
    path = Path('/place/with/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']
    key = 'new.key'

    a = Artifact(path, filter_terms)
    initial_keys = a.keys

    assert key not in a.keys

    with pytest.raises(ArtifactException):
        a.write(key, None)

    assert key not in a.keys
    assert key not in a._cache
    hdf_mock.write.called_once_with(path, 'metadata.keyspace', ['metadata.keyspace'])
    hdf_mock.remove.assert_not_called()
    assert a.keys == initial_keys


def test_artifact_write(hdf_mock, keys_mock):
    path = Path(Path('/place/with/artifact.hdf'))
    filter_terms = ['location == Global', 'draw == 10']
    key = 'new.key'

    a = Artifact(path, filter_terms)
    initial_keys = a.keys

    assert key not in a.keys

    a.write(key, "data")

    assert key in a.keys
    assert key not in a._cache
    expected_call = [call(path, 'metadata.keyspace', ['metadata.keyspace']),
                     call(path, 'metadata.keyspace', keys_mock + [key]),
                     call(path, key, 'data')]
    assert hdf_mock.write.call_args_list == expected_call
    assert set(a.keys) == set(initial_keys + [key])


def test_remove_bad_key(hdf_mock):
    path = Path('/place/with/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']
    key = 'non_existent.key'
    a = Artifact(path, filter_terms)
    initial_keys = a.keys

    assert key not in a.keys
    assert key not in a._cache

    with pytest.raises(ArtifactException) as err_info:
        a.remove(key)

    assert f'Trying to remove non-existent key {key} from artifact.' == str(err_info.value)
    assert key not in a.keys
    assert key not in a._cache
    hdf_mock.remove.assert_not_called()
    hdf_mock.write.called_once_with(path, 'metadata.keyspace', ['metadata.keyspace'])
    assert a.keys == initial_keys


def test_remove_no_cache(hdf_mock, keys_mock):
    path = Path('/place/with/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']
    key = 'population.structure'

    a = Artifact(path, filter_terms)

    initial_keys = a.keys[:]

    assert key in initial_keys
    assert key not in a._cache

    a.remove(key)

    assert key not in a.keys
    assert key not in a._cache
    assert set(initial_keys).difference(a.keys) == {key}
    expected_calls_remove = [call(path, 'metadata.keyspace'), call(path, key)]
    assert hdf_mock.remove.call_args_list == expected_calls_remove
    expected_calls_write = [call(path, 'metadata.keyspace', ['metadata.keyspace']),
                            call(path, 'metadata.keyspace', [k for k in keys_mock if k != key])]
    assert hdf_mock.write.call_args_list == expected_calls_write


def test_remove(hdf_mock):
    path = Path('/place/with/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']
    key = 'population.structure'

    a = Artifact(path, filter_terms)
    a._cache[key] = 'data'

    assert key in a.keys
    assert key in a._cache

    a.remove(key)

    assert key not in a.keys
    assert key not in a._cache

    expected_calls = [call(path, 'metadata.keyspace'), call(path, key)]
    assert hdf_mock.remove.call_args_list == expected_calls


def test_clear_cache(hdf_mock):
    path = Path('/place/with/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']
    key = 'population.structure'

    a = Artifact(path, filter_terms)
    a.clear_cache()

    assert a._cache == {}

    a._cache[key] = 'data'
    a.clear_cache()

    assert a._cache == {}


def test_loading_key_leaves_filters_unchanged(hdf_mock):
    # loading each key will drop the fake_filter from filter_terms for that key
    # make sure that artifact's filter terms stay the same though
    path = str(Path(__file__).parent / 'artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10', 'fake_filter']

    a = Artifact(path, filter_terms=filter_terms)

    for key in _KEYS:
        a.load(key)
        assert a.filter_terms == filter_terms


def test_replace(hdf_mock, keys_mock):
    path = Path('/place/with/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']
    key = 'new.key'

    a = Artifact(path, filter_terms=filter_terms)

    assert key not in a.keys

    a.write(key, "data")
    keyspace_key = 'metadata.keyspace'
    new_keyspace = [k for k in keys_mock + [key]]

    assert hdf_mock.write.call_args_list == [call(path, keyspace_key, [str(keyspace_key)]),
                                             call(path, keyspace_key, new_keyspace),
                                             call(path, key, 'data')]

    hdf_mock.reset_mock()

    a.replace(key, "new_data")

    # keyspace will be remove first in self.remove from a.replace then self.write from a.replace
    expected_calls_remove = [call(path, keyspace_key), call(path, key), call(path, keyspace_key)]
    assert hdf_mock.remove.call_args_list == expected_calls_remove

    expected_calls_write = [call(path, keyspace_key, new_keyspace),
                            call(path, keyspace_key, new_keyspace), call(path, key, 'new_data')]
    assert hdf_mock.write.call_args_list == expected_calls_write
    assert key in a.keys


def test_replace_nonexistent_key(hdf_mock):
    path = Path('/place/with/artifact.hdf')
    filter_terms = ['location == Global', 'draw == 10']
    key = 'new.key'

    a = Artifact(path, filter_terms=filter_terms)
    hdf_mock.called_once_with(key)
    assert key not in a.keys

    hdf_mock.reset_mock()
    with pytest.raises(ArtifactException):
        a.replace(key, "new_data")

    hdf_mock.write.assert_not_called()
    hdf_mock.remove.assert_not_called()


def test_to_tree():
    keys = ['population.structure',
            'population.age_groups',
            'cause.diarrheal_diseases.incidence',
            'cause.diarrheal_diseases.prevalence']

    key_tree = {
        'population': {
            'structure': [],
            'age_groups': [],
        },
        'cause': {
            'diarrheal_diseases': ['incidence', 'prevalence']
        }
    }

    assert _to_tree(keys) == key_tree


def test_create_hdf(tmpdir):
    path = Path(tmpdir)/'test.hdf'
    assert not path.is_file()

    test_artifact = Artifact(path)
    assert path.is_file()
    assert 'metadata.keyspace' in test_artifact

    test_artifact.write('new.key', 'data')
    assert 'new.key' in test_artifact

    #  check whether the existing path was NOT wiped out
    new_artifact = Artifact(test_artifact.path)
    assert new_artifact.path == test_artifact.path
    assert 'new.key' in new_artifact


def test_keys_initialization(tmpdir):
    path = Path(tmpdir)/'test.hdf'
    test_artifact = Artifact(path)
    test_key = test_artifact._keys

    assert test_artifact._path == test_key._path
    assert test_key._keys == ['metadata.keyspace']

    test_artifact.write('new.keys', 'data')
    assert test_key._keys == ['metadata.keyspace', 'new.keys']
    assert test_key.to_list() == test_artifact.keys


def test_keys_append(tmpdir):
    path = Path(tmpdir)/'test.hdf'
    test_artifact = Artifact(path)
    test_keys = test_artifact._keys

    test_artifact.write('test.keys', 'data')
    assert 'test.keys' in test_artifact
    assert 'test.keys' in test_keys
    assert test_keys._keys == ['metadata.keyspace', 'test.keys'] == [str(k) for k in test_artifact.keys]


def test_keys_remove(tmpdir):
    path = Path(tmpdir) / 'test.hdf'
    test_artifact = Artifact(path)
    test_keys = test_artifact._keys

    test_artifact.write('test.keys1', 'data')
    test_artifact.write('test.keys2', 'data')
    assert 'test.keys1' in test_artifact and 'test.keys2' in test_artifact
    assert 'test.keys1' in test_keys and 'test.keys2' in test_keys

    test_artifact.remove('test.keys2')
    assert 'test.keys1' in test_artifact and 'test.keys2' not in test_artifact
    assert 'test.keys1' in test_keys and 'test.keys2' not in test_keys


@pytest.mark.parametrize('filters, error, match', [(['draw == 0 & year > 2011', 'age < 2015 | draw in [1, 2, 3]'],
                                                    ValueError, 'only supply one'),
                                                   (['year > 2010 & age < 5 & parameter == "cat1" & '
                                                     'draw==0 | draw==100'],
                                                    ValueError, 'only supply one'),
                                                   (['draw >= 10'], NotImplementedError, 'only supported'),
                                                   (['draw<5'], NotImplementedError, 'only supported')])
def test__parse_draw_filters_fail(filters, error, match):
    with pytest.raises(error, match=match):
        _parse_draw_filters(filters)


@pytest.mark.parametrize('draw_operator, draw_values', [('=', [5]),
                                                        ('==', [10]),
                                                        (' = ', [100]),
                                                        (' == ', [12]),
                                                        (' in ', [1, 7, 160]),
                                                        ('    in            ', [140, 2, 14])])
def test__parse_draw_filters_pass(draw_operator, draw_values):
    draw_filter = f'draw{draw_operator}{draw_values if "in" in draw_operator else draw_values[0]}'
    expected_cols = [f'draw_{d}' for d in draw_values] + ['value']

    assert _parse_draw_filters([draw_filter]) == expected_cols

    complicated_filter = [f'year > 2010 & age < 12', 'other_col in [1, 2, 40]', 'one_more > 10']
    for i in range(len(complicated_filter)):
        filters = complicated_filter.copy()
        filters[i] += " | " + draw_filter
        assert _parse_draw_filters(filters) == expected_cols

    assert _parse_draw_filters(complicated_filter + [draw_filter]) == expected_cols

    assert _parse_draw_filters([]) is None



import pytest
import pandas as pd
import numpy as np

import vivarium.framework.randomness as random
from vivarium.framework.randomness import RandomnessManager, RandomnessStream, RESIDUAL_CHOICE, RandomnessError


@pytest.fixture(params=[10**4, 10**5])
def index(request):
    return pd.Index(range(request.param)) if request.param else None


@pytest.fixture(params=[['a', 'small', 'bird']])
def choices(request):
    return request.param


# TODO: Add 2-d weights to the tests.
@pytest.fixture(params=[None, [10, 10, 10], [0.5, 0.1, 0.4]])
def weights(request):
    return request.param


@pytest.fixture(params=[(0.2, 0.1, RESIDUAL_CHOICE),
                        (0.5, 0.6, RESIDUAL_CHOICE),
                        (.1, RESIDUAL_CHOICE, RESIDUAL_CHOICE)])
def weights_with_residuals(request):
    return request.param


def test_normalize_shape(weights_with_residuals, index):
    p = random._normalize_shape(weights_with_residuals, index)
    assert p.shape == (len(index), len(weights_with_residuals))


def test__set_residual_probability(weights_with_residuals, index):
    # Coerce the weights to a 2-d numpy array.
    p = random._normalize_shape(weights_with_residuals, index)

    residual = np.where(p == RESIDUAL_CHOICE, 1, 0)
    non_residual = np.where(p != RESIDUAL_CHOICE, p, 0)

    if np.any(non_residual.sum(axis=1) > 1):
        with pytest.raises(RandomnessError):
            # We received un-normalized probability weights.
            random._set_residual_probability(p)

    elif np.any(residual.sum(axis=1) > 1):
        with pytest.raises(RandomnessError):
            # We received multiple instances of `RESIDUAL_CHOICE`
            random._set_residual_probability(p)

    else:  # Things should work
        p_total = np.sum(random._set_residual_probability(p))
        assert np.isclose(p_total, len(index), atol=0.0001)


def test_filter_for_probability(index):
    dates = [pd.Timestamp(1991, 1, 1), pd.Timestamp(1990, 1, 1)]
    randomness = RandomnessStream('test', dates.pop, 1)

    sub_index = randomness.filter_for_probability(index, 0.5)
    assert round(len(sub_index)/len(index), 1) == 0.5

    sub_sub_index = randomness.filter_for_probability(sub_index, 0.5)
    assert round(len(sub_sub_index)/len(sub_index), 1) == 0.5


def test_choice(index, choices, weights):
    dates = [pd.Timestamp(1990, 1, 1)]
    randomness = RandomnessStream('test', dates.pop, 1)

    chosen = randomness.choice(index, choices, p=weights)
    count = chosen.value_counts()
    # If we have weights, normalize them, otherwise generate uniform weights.
    weights = [w/sum(weights) for w in weights] if weights else [1/len(choices) for _ in choices]
    for k, c in count.items():
        assert np.isclose(c/len(index), weights[choices.index(k)], atol=0.01)


def test_choice_with_residuals(index, choices, weights_with_residuals):
    print(RESIDUAL_CHOICE in weights_with_residuals)
    dates = [pd.Timestamp(1990, 1, 1)]
    randomness = RandomnessStream('test', dates.pop, 1)

    p = random._normalize_shape(weights_with_residuals, index)

    residual = np.where(p == RESIDUAL_CHOICE, 1, 0)
    non_residual = np.where(p != RESIDUAL_CHOICE, p, 0)

    if np.any(non_residual.sum(axis=1) > 1):
        with pytest.raises(RandomnessError):
            # We received un-normalized probability weights.
            randomness.choice(index, choices, p=weights_with_residuals)

    elif np.any(residual.sum(axis=1) > 1):
        with pytest.raises(RandomnessError):
            # We received multiple instances of `RESIDUAL_CHOICE`
            randomness.choice(index, choices, p=weights_with_residuals)

    else:  # Things should work
        chosen = randomness.choice(index, choices, p=weights_with_residuals)
        count = chosen.value_counts()
        print(weights_with_residuals)
        # We're relying on the fact that weights_with_residuals is a 1-d list
        residual_p = 1 - sum([w for w in weights_with_residuals if w != RESIDUAL_CHOICE])
        weights = [w if w != RESIDUAL_CHOICE else residual_p for w in weights_with_residuals]

        for k, c in count.items():
            assert np.isclose(c / len(index), weights[choices.index(k)], atol=0.01)


def mock_clock():
    return pd.Timestamp('1/1/2005')


def test_RandomnessManager_get_randomness_stream():
    seed = 123456

    rm = RandomnessManager()
    rm._add_constraint = lambda f, **kwargs: f
    rm._seed = seed
    rm._clock = mock_clock
    stream = rm._get_randomness_stream('test')

    assert stream.key == 'test'
    assert stream.seed == seed
    assert stream.clock is mock_clock
    assert set(rm._decision_points.keys()) == {'test'}

    with pytest.raises(RandomnessError):
        rm.get_randomness_stream('test')


def test_RandomnessManager_register_simulants():
    seed = 123456
    rm = RandomnessManager()
    rm._add_constraint = lambda f, **kwargs: f
    rm._seed = seed
    rm._clock = mock_clock
    rm._key_columns = ['age', 'sex']

    bad_df = pd.DataFrame({'age': range(10),
                           'not_sex': [1]*5 + [2]*5})
    with pytest.raises(RandomnessError):
        rm.register_simulants(bad_df)

    good_df = pd.DataFrame({'age': range(10),
                            'sex': [1]*5 + [2]*5})

    rm.register_simulants(good_df)
    assert rm._key_mapping._map.index.difference(good_df.set_index(good_df.columns.tolist()).index).empty



"""
====================
The Simulation Clock
====================

The components here provide implementations of different kinds of simulation
clocks for use in ``vivarium``.

For more information about time in the simulation, see the associated
:ref:`concept note <time_concept>`.

"""
from typing import Union, Callable
from numbers import Number
from datetime import datetime, timedelta

import pandas as pd


Time = Union[datetime, Number]
Timedelta = Union[timedelta, Number]


class SimulationClock:
    """Defines a base implementation for a simulation clock."""

    def __init__(self):
        self._time = None
        self._stop_time = None
        self._step_size = None

    @property
    def name(self):
        return "simulation_clock"

    @property
    def time(self) -> Time:
        """The current simulation time."""
        assert self._time is not None, 'No start time provided'
        return self._time

    @property
    def stop_time(self) -> Time:
        """The time at which the simulation will stop."""
        assert self._stop_time is not None, 'No stop time provided'
        return self._stop_time

    @property
    def step_size(self) -> Timedelta:
        """The size of the next time step."""
        assert self._step_size is not None, 'No step size provided'
        return self._step_size

    def step_forward(self) -> None:
        """Advances the clock by the current step size."""
        self._time += self.step_size

    def step_backward(self):
        """Rewinds the clock by the current step size."""
        self._time -= self.step_size


class SimpleClock(SimulationClock):
    """A unitless step-count based simulation clock."""

    configuration_defaults = {
        'time': {
            'start': 0,
            'end': 100,
            'step_size': 1,
        }
    }

    @property
    def name(self):
        return "simple_clock"

    def setup(self, builder):
        self._time = builder.configuration.time.start
        self._stop_time = builder.configuration.time.end
        self._step_size = builder.configuration.time.step_size

    def __repr__(self):
        return "SimpleClock()"


def get_time_stamp(time):
    return pd.Timestamp(time['year'], time['month'], time['day'])


class DateTimeClock(SimulationClock):
    """A date-time based simulation clock."""

    configuration_defaults = {
        'time': {
            'start': {
                'year': 2005,
                'month': 7,
                'day': 2
            },
            'end': {
                'year': 2010,
                'month': 7,
                'day': 2,
            },
            'step_size': 1,  # Days
        }
    }

    @property
    def name(self):
        return "datetime_clock"

    def setup(self, builder):
        time = builder.configuration.time
        self._time = get_time_stamp(time.start)
        self._stop_time = get_time_stamp(time.end)
        self._step_size = pd.Timedelta(days=time.step_size // 1, hours=(time.step_size % 1) * 24)

    def __repr__(self):
        return "DateTimeClock()"


class TimeInterface:
    def __init__(self, manager: SimulationClock):
        self._manager = manager

    def clock(self) -> Callable[[], Union[datetime, Number]]:
        """Gets a callable that returns the current simulation time."""
        return lambda: self._manager.time

    def step_size(self) -> Callable[[], Union[timedelta, Number]]:
        """Gets a callable that returns the current simulation step size."""
        return lambda: self._manager.step_size



"""
==============
Output Metrics
==============

Currently, ``vivarium`` uses the :ref:`values pipeline system <values_concept>`
to produce the results, or output metrics, from a simulation.  The metrics
The component here is a normal ``vivarium`` component whose only purpose is
to provide an empty :class:`dict` as the source of the *"Metrics"* pipeline.
It is included by default in all simulations.
"""


class Metrics:
    """This class declares a value pipeline that allows other components to store summary metrics."""

    @property
    def name(self):
        return "metrics"

    def setup(self, builder):
        self.metrics = builder.value.register_value_producer('metrics', source=lambda index: {})

    def __repr__(self):
        return "Metrics()"



import pytest

from vivarium.framework.lifecycle import (LifeCycleState, LifeCyclePhase, LifeCycle,
                                          LifeCycleError, ConstraintError,
                                          LifeCycleManager, ConstraintMaker)


def test_state_add_next():
    a = LifeCycleState('a')
    b = LifeCycleState('b')
    c = LifeCycleState('c')

    assert a.valid_next_state(None)
    a.add_next(b)
    assert a.valid_next_state(b)
    assert not a.valid_next_state(c)
    assert not a.valid_next_state(None)

    assert b.valid_next_state(None)
    b.add_next(a, loop=True)
    assert b.valid_next_state(a)
    assert b.valid_next_state(None)
    b.add_next(c)
    assert not b.valid_next_state(None)
    assert b.valid_next_state(a)
    assert b.valid_next_state(c)

    assert c.valid_next_state(None)
    assert not c.valid_next_state(a)
    assert not c.valid_next_state(b)


def test_state_entrance_count():
    a = LifeCycleState('a')
    assert a.entrance_count == 0
    a.enter()
    a.enter()
    assert a.entrance_count == 2


class BumbleBee:

    def __init__(self, name):
        self.name = name

    def buzz(self):
        return "Buzzzzz"

    def what_we_doin(self):
        return "Gettin' honey."

    def __call__(self):
        return  "On the way!"


def test_state_add_handlers():
    a = LifeCycleState('a')
    b = BumbleBee('Alice')
    a.add_handlers([b.buzz, b.what_we_doin, test_state_add_handlers])
    assert a._handlers == ['BumbleBee(Alice).buzz', 'BumbleBee(Alice).what_we_doin',
                           'Unbound function test_state_add_handlers']


def test_phase_init_no_loop():
    phase_name = 'test'
    state_names = ['a', 'b', 'c', 'd', 'e']

    p = LifeCyclePhase(phase_name, state_names, loop=False)
    assert p.name == phase_name
    assert [s.name for s in p.states] == state_names
    current = p.states[0]
    for state in p.states[1:]:
        assert current.valid_next_state(state)
        current = state
    assert current.valid_next_state(None)


def test_phase_init_loop():
    phase_name = 'test'
    state_names = ['a', 'b', 'c', 'd', 'e']

    p = LifeCyclePhase(phase_name, state_names, loop=True)
    assert p.name == phase_name
    assert [s.name for s in p.states] == state_names
    current = p.states[0]
    for state in p.states[1:]:
        assert current.valid_next_state(state)
        current = state
    assert current.valid_next_state(p.states[0])


def test_phase_add_next():
    phase1_name = 'test1'
    phase2_name = 'test2'
    states1 = ['a', 'b', 'c', 'd', 'e']
    states2 = ['1', '2', '3', '4', '5']

    p1 = LifeCyclePhase(phase1_name, states1, loop=True)
    p2 = LifeCyclePhase(phase2_name, states2, loop=False)

    p1.add_next(p2)
    assert p1.states[-1].valid_next_state(p1.states[0])
    assert p1.states[-1].valid_next_state(p2.states[0])


def test_phase_get_state():
    name = 'test'
    state_names = ['a', 'b', 'c', 'd', 'e']

    p = LifeCyclePhase(name, state_names, loop=False)

    for name, state in zip(state_names, p.states):
        assert state is p.get_state(name)

    with pytest.raises(IndexError):
        p.get_state('not_a_state')


def test_phase_contains():
    name = 'test'
    state_names = ['a', 'b', 'c', 'd', 'e']

    p = LifeCyclePhase(name, state_names, loop=False)

    for name in state_names:
        assert name in p

    for name in ['1', '2', '3', '']:
        assert name not in p


def test_lifecycle_init():
    lc = LifeCycle()
    assert 'initialization' in lc
    assert 'not_a_state' not in lc


def test_lifecycle_validate():
    lc = LifeCycle()

    with pytest.raises(LifeCycleError, match='phase names must be unique'):
        lc._validate('initialization', ['a', 'b', 'c'])

    with pytest.raises(LifeCycleError, match='duplicate state names'):
        lc._validate('new_phase', ['a', 'b', 'c', 'b'])

    with pytest.raises(LifeCycleError, match='initialization'):
        lc._validate('new_phase', ['a', 'b', 'initialization'])

    lc._validate('new_phase', ['a', 'b', 'c'])


@pytest.mark.parametrize('loop', [True, False])
def test_lifecycle_add_phase_fail(loop):
    lc = LifeCycle()

    with pytest.raises(LifeCycleError, match='phase names must be unique'):
        lc.add_phase('initialization', ['a', 'b', 'c'], loop)

    with pytest.raises(LifeCycleError, match='duplicate state names'):
        lc.add_phase('new_phase', ['a', 'b', 'c', 'b'], loop)

    with pytest.raises(LifeCycleError, match='initialization'):
        lc.add_phase('new_phase', ['a', 'b', 'initialization'], loop)


@pytest.mark.parametrize('loop', [True, False])
def test_lifecycle_add_phase(loop):
    lc = LifeCycle()
    lc.add_phase('phase1', ['a', 'b', 'c'], loop)


def test_lifecycle_get_state():
    lc = LifeCycle()
    assert lc.get_state('initialization').name == 'initialization'

    with pytest.raises(LifeCycleError, match='non-existent state'):
        lc.get_state('not_a_state')


def test_lifecycle_get_states():
    lc = LifeCycle()
    assert lc.get_state_names('initialization') == ['initialization']

    with pytest.raises(LifeCycleError, match='non-existent phase'):
        lc.get_state_names('not_a_phase')


def test_lifecycle_integration():
    lc = LifeCycle()

    with pytest.raises(LifeCycleError, match='non-existent state'):
        lc.get_state('a')
    with pytest.raises(LifeCycleError, match='non-existent phase'):
        lc.get_state_names('phase1')

    init = lc.get_state('initialization')
    assert init.valid_next_state(None)

    lc.add_phase('phase1', ['a'], loop=False)
    a = lc.get_state('a')

    assert lc.get_state_names('phase1') == ['a']
    assert init.valid_next_state(a)
    assert not init.valid_next_state(None)
    assert a.valid_next_state(None)

    lc.add_phase('phase2', ['b'], loop=True)
    b = lc.get_state('b')

    assert lc.get_state_names('phase2') == ['b']
    assert a.valid_next_state(b)
    assert b.valid_next_state(None)
    assert b.valid_next_state(b)

    lc.add_phase('phase3', ['c', 'd'], loop=True)
    c = lc.get_state('c')
    d = lc.get_state('d')

    assert not b.valid_next_state(None)
    assert b.valid_next_state(b)
    assert b.valid_next_state(c)
    assert c.valid_next_state(d)
    assert d.valid_next_state(c)
    assert d.valid_next_state(None)

    with pytest.raises(LifeCycleError, match='phase names must be unique'):
        lc.add_phase('phase2', ['1', '2'], loop=True)

    with pytest.raises(LifeCycleError, match='duplicate state names'):
        lc.add_phase('phase4', ['1', '2', '3', '2'], loop=False)

    with pytest.raises(LifeCycleError, match='state names must be unique'):
        lc.add_phase('phase4', ['a', 'f', 'g'], loop=False)

    with pytest.raises(LifeCycleError, match='non-existent state'):
        lc.get_state('not a state')

    with pytest.raises(LifeCycleError, match='non-existent phase'):
        lc.get_state_names('not a phase')


def test_manager_init():
    lm = LifeCycleManager()
    assert lm.name == 'life_cycle_manager'
    assert lm.current_state == 'initialization'


@pytest.mark.parametrize('loop', [True, False])
def test_lifecycle_manager_add_phase_fail(loop):
    lm = LifeCycleManager()

    with pytest.raises(LifeCycleError, match='phase names must be unique'):
        lm.add_phase('initialization', ['a', 'b', 'c'], loop)

    with pytest.raises(LifeCycleError, match='duplicate state names'):
        lm.add_phase('new_phase', ['a', 'b', 'c', 'b'], loop)

    with pytest.raises(LifeCycleError, match='initialization'):
        lm.add_phase('new_phase', ['a', 'b', 'initialization'], loop)


@pytest.mark.parametrize('loop', [True, False])
def test_lifecycle_manager_add_phase(loop):
    lm = LifeCycleManager()
    lm.add_phase('phase1', ['a', 'b', 'c'], loop)


def test_lifecycle_manager_get_states():
    lm = LifeCycleManager()
    assert lm.get_state_names('initialization') == ['initialization']

    with pytest.raises(LifeCycleError, match='non-existent phase'):
        lm.get_state_names('not_a_phase')


def test_lifecycle_manager_set_state():
    lm = LifeCycleManager()
    lm.add_phase('phase1', ['a', 'b'])
    lm.add_phase('phase2', ['c', 'd'])

    assert lm.current_state == 'initialization'
    lm.set_state('a')
    assert lm.current_state == 'a'
    lm.set_state('b')
    assert lm.current_state == 'b'

    with pytest.raises(LifeCycleError, match='Invalid transition'):
        lm.set_state('a')  # phase1 does not loop

    with pytest.raises(LifeCycleError, match='Invalid transition'):
        lm.set_state('d')  # next state is c

    with pytest.raises(LifeCycleError, match='non-existent state'):
        lm.set_state('e')

    lm.set_state('c')
    assert lm.current_state == 'c'
    lm.set_state('d')
    assert lm.current_state == 'd'


def test_lifecycle_manager_set_state_with_loop():
    lm = LifeCycleManager()
    lm.add_phase('phase1', ['a', 'b'], loop=True)
    lm.add_phase('phase2', ['c', 'd'])

    assert lm.current_state == 'initialization'
    lm.set_state('a')
    assert lm.current_state == 'a'
    lm.set_state('b')
    assert lm.current_state == 'b'

    lm.set_state('a')  # phase permits loops
    assert lm.current_state == 'a'
    lm.set_state('b')
    assert lm.current_state == 'b'

    with pytest.raises(LifeCycleError, match='Invalid transition'):
        lm.set_state('d')  # next state is a or c

    with pytest.raises(LifeCycleError, match='non-existent state'):
        lm.set_state('e')

    lm.set_state('c')
    assert lm.current_state == 'c'
    lm.set_state('d')
    assert lm.current_state == 'd'

    with pytest.raises(LifeCycleError, match='Invalid transition'):
        lm.set_state('c')  # phase 2 does not permit loops


def test_lifecycle_manager_add_handlers():
    lm = LifeCycleManager()
    lm.add_phase('phase1', ['a'])

    init = lm.lifecycle.get_state('initialization')
    a = lm.lifecycle.get_state('a')

    b = BumbleBee('Alice')

    lm.add_handlers('initialization', [b.buzz, b.what_we_doin, test_lifecycle_manager_add_handlers])
    assert init._handlers == ['BumbleBee(Alice).buzz',
                              'BumbleBee(Alice).what_we_doin',
                              'Unbound function test_lifecycle_manager_add_handlers']

    lm.add_handlers('a', [b.buzz])
    assert a._handlers == ['BumbleBee(Alice).buzz']


def test_lifecycle_manager_add_constraint_fail():
    lm = LifeCycleManager()
    lm.add_phase('phase1', ['a', 'b', 'c'])

    alice = BumbleBee('Alice')

    with pytest.raises(ValueError, match='Must provide exactly one'):
        lm.add_constraint(alice.buzz)

    with pytest.raises(ValueError, match='Must provide exactly one'):
        lm.add_constraint(alice.buzz, allow_during=['initialization', 'a'], restrict_during=['c'])

    with pytest.raises(LifeCycleError, match='states not in the life cycle'):
        lm.add_constraint(alice.buzz, allow_during=['not_a_state'])

    with pytest.raises(LifeCycleError, match='states not in the life cycle'):
        lm.add_constraint(alice.buzz, restrict_during=['not_a_state'])

    lm.add_constraint(alice.buzz, allow_during=['a'])

    with pytest.raises(ConstraintError, match='already been constrained'):
        lm.add_constraint(alice.buzz, allow_during=['b'])

    with pytest.raises(TypeError, match='bound object methods'):
        lm.add_constraint(test_lifecycle_manager_add_constraint_fail, allow_during=['a'])

    with pytest.raises(ValueError, match='normal object methods'):
        lm.add_constraint(alice.__call__, allow_during=['a'])


def test_lifecycle_manager_add_constraint():
    lm = LifeCycleManager()
    lm.add_phase('phase1', ['a', 'b', 'c'], loop=True)

    alice = BumbleBee('Alice')
    bob = BumbleBee('Bob')

    lm.add_constraint(alice.buzz, allow_during=['initialization', 'a'])
    lm.add_constraint(bob.buzz, restrict_during=['initialization', 'c'])

    alice.buzz()
    with pytest.raises(ConstraintError, match='it may only be called during'):
        bob.buzz()

    lm.set_state('a')
    alice.buzz()
    bob.buzz()

    lm.set_state('b')
    with pytest.raises(ConstraintError, match='it may only be called during'):
        alice.buzz()
    bob.buzz()

    lm.set_state('c')
    with pytest.raises(ConstraintError, match='it may only be called during'):
        alice.buzz()
    with pytest.raises(ConstraintError, match='it may only be called during'):
        bob.buzz()

    lm.set_state('a')
    alice.buzz()
    bob.buzz()



from .hdf import EntityKey
from .artifact import Artifact, ArtifactException
from .manager import (ArtifactManager, ArtifactInterface, parse_artifact_path_config,
                      get_location_term, filter_data, validate_filter_term)



"""
==========
Exceptions
==========

Module containing framework-wide exception definitions. Exceptions for
particular subsystems are defined in their respective modules.

"""


class VivariumError(Exception):
    """Generic exception raised for errors in ``vivarium`` simulations."""
    pass



from itertools import chain, combinations, product

import pytest
import numpy as np
import pandas as pd
from scipy.stats import chisquare

from vivarium.framework.randomness import IndexMap, RandomnessError


def almost_powerset(iterable):
    """almost_powerset([1,2,3]) --> (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    return list(chain.from_iterable(combinations(s, r) for r in range(1, len(s)+1)))


def generate_keys(number, types=('int', 'float', 'datetime'), seed=123456):
    rs = np.random.RandomState(seed=seed)

    index = []
    if 'datetime' in types:
        year = rs.choice(np.arange(1980, 2018))
        day = rs.choice(pd.date_range(f'01/01/{year}', periods=365))
        start_time = rs.choice(pd.date_range(day, periods=86400, freq='s'))
        freq = rs.choice(['ms', 's', 'min', 'h'])
        index.append(pd.date_range(start_time, periods=number, freq=freq))

    if 'int' in types:
        kind = rs.choice(['random', 'sequential'])
        if kind == 'random':
            ints = np.unique(rs.randint(0, 1000 * number, size=100 * number))
            assert len(ints) > number
            rs.shuffle(ints)
            index.append(ints[:number])
        else:
            start = rs.randint(0, 100*number)
            index.append(np.arange(start, start + number, dtype=int))

    if 'float' in types:
        index.append(rs.random_sample(size=number))

    return pd.Series(0, index=index).index


rs = np.random.RandomState(seed=456789)
index_sizes = list(rs.randint(10_000, 250_000, size=1))
types = list(almost_powerset(['int', 'float', 'datetime']))
seeds = list(rs.randint(10000000, size=1))


def id_fun(param):
    return f"Size:{param[0]}, Types:{param[1]}, Seed:{param[2]}"


@pytest.fixture(scope='module', params=list(product(index_sizes, types, seeds)), ids=id_fun)
def map_size_and_hashed_values(request):
    keys = generate_keys(*request.param)
    m = IndexMap()
    return m.map_size, m.hash_(keys)


def test_digit_scalar():
    m = IndexMap()
    k = 123456789
    for i in range(10):
        assert m.digit(k, i) == 10 - (i + 1)


def test_digit_series():
    m = IndexMap()
    k = pd.Series(123456789, index=range(10000))
    for i in range(10):
        assert len(m.digit(k, i).unique()) == 1
        assert m.digit(k, i)[0] == 10 - (i + 1)


def test_clip_to_seconds_scalar():
    m = IndexMap()
    k = pd.to_datetime("2010-01-25 06:25:31.123456789")
    assert m.clip_to_seconds(k.value) == int(str(k.value)[:10])


def test_clip_to_seconds_series():
    m = IndexMap()
    stamp = 1234567890
    k = pd.date_range(pd.to_datetime(stamp, unit='s'), periods=10000, freq='ns').to_series().astype(np.int64)
    assert len(m.clip_to_seconds(k).unique()) == 1
    assert m.clip_to_seconds(k).unique()[0] == stamp


def test_spread_scalar():
    m = IndexMap()
    assert m.spread(1234567890) == 4072825790


def test_spread_series():
    m = IndexMap()
    s = pd.Series(1234567890, index=range(10000))
    assert len(m.spread(s).unique()) == 1
    assert m.spread(s).unique()[0] == 4072825790


def test_shift_scalar():
    m = IndexMap()
    assert m.shift(1.1234567890) == 1234567890


def test_shift_series():
    m = IndexMap()
    s = pd.Series(1.1234567890, index=range(10000))
    assert len(m.shift(s).unique()) == 1
    assert m.shift(s).unique()[0] == 1234567890


def test_convert_to_ten_digit_int():
    m = IndexMap()
    v = 1234567890
    datetime_col = pd.date_range(pd.to_datetime(v, unit='s'), periods=10000, freq='ns').to_series()
    int_col = pd.Series(v, index=range(10000))
    float_col = pd.Series(1.1234567890, index=range(10000))
    bad_col = pd.Series('a', index=range(10000))

    assert m.convert_to_ten_digit_int(datetime_col).unique()[0] == v
    assert m.convert_to_ten_digit_int(int_col).unique()[0] == 4072825790
    assert m.convert_to_ten_digit_int(float_col).unique()[0] == v
    with pytest.raises(RandomnessError):
        m.convert_to_ten_digit_int(bad_col)


@pytest.mark.skip("This fails because the hash needs work")
def test_hash_collisions(map_size_and_hashed_values):
    n, h = map_size_and_hashed_values
    k = len(h)

    expected_empty_bins = n*(1 - 1/n)**k  # Potential source of roundoff issues.
    expected_full_bins = n - expected_empty_bins
    expected_collisions = k - expected_full_bins

    uniques = h.drop_duplicates()
    actual_collisions = len(h) - len(uniques)

    assert actual_collisions/expected_collisions < 1.5, "Too many collisions"


@pytest.mark.skip("This fails because the hash needs work")
def test_hash_uniformity(map_size_and_hashed_values):
    n, h = map_size_and_hashed_values

    k = len(h)
    num_bins = k//5  # Want about 5 items per bin for chi-squared
    bins = np.linspace(0, n + 1, num_bins)

    binned_data = pd.cut(h, bins)
    distribution = pd.value_counts(binned_data).sort_index()
    c, p = chisquare(distribution)

    assert p > 0.05, "Data not uniform"


def test_update(mocker):
    m = IndexMap()
    keys = generate_keys(10000)

    def hash_mock(k, salt=0):
        seed = 123456
        rs = np.random.RandomState(seed=seed + salt)
        return pd.Series(rs.randint(0, len(k)*10, size=len(k)), index=k)

    mocker.patch.object(m, 'hash_', side_effect=hash_mock)
    m.update(keys)
    assert len(m) == len(keys), "All keys not in mapping"
    assert m._map.index.difference(keys).empty, "All keys not in mapping"
    assert len(m._map.unique()) == len(keys), "Duplicate values in mapping"

    # Can't have duplicate keys.
    with pytest.raises(KeyError):
        m.update(keys)

    new_unique_keys = generate_keys(1000).difference(keys)
    m.update(new_unique_keys)
    assert len(m) == len(keys) + len(new_unique_keys), "All keys not in mapping"
    assert m._map.index.difference(keys.union(new_unique_keys)).empty, "All keys not in mapping"
    assert len(m._map.unique()) == len(keys) + len(new_unique_keys), "Duplicate values in mapping"



import pandas as pd


class Risk:

    configuration_defaults = {
        'risk': {
            'proportion_exposed': 0.3,
        },
    }

    def __init__(self, name):
        self.name = name
        self.configuration_defaults = {name: Risk.configuration_defaults['risk']}

    def setup(self, builder):
        proportion_exposed = builder.configuration[self.name].proportion_exposed
        self.base_exposure_threshold = builder.value.register_value_producer(
            f'{self.name}.base_proportion_exposed', source=lambda index: pd.Series(proportion_exposed, index=index)
        )
        self.exposure_threshold = builder.value.register_value_producer(
            f'{self.name}.proportion_exposed', source=self.base_exposure_threshold
        )

        self.exposure = builder.value.register_value_producer(f'{self.name}.exposure', source=self._exposure)
        self.randomness = builder.randomness.get_stream(self.name)

        columns_created = [f'{self.name}_propensity']
        builder.population.initializes_simulants(self.on_initialize_simulants,
                                                 creates_columns=columns_created,
                                                 requires_streams=[self.name])
        self.population_view = builder.population.get_view(columns_created)

    def on_initialize_simulants(self, pop_data):
        draw = self.randomness.get_draw(pop_data.index)
        self.population_view.update(pd.Series(draw, name=f'{self.name}_propensity'))

    def _exposure(self, index):
        propensity = self.population_view.get(index)[f'{self.name}_propensity']
        return self.exposure_threshold(index) > propensity

    def __repr__(self):
        return f"Risk(name={self.name})"


class DirectEffect:

    configuration_defaults = {
        'direct_effect': {
            'relative_risk': 2,
        },
    }

    def __init__(self, risk, disease_rate):
        self.risk = risk
        self.disease_rate = disease_rate
        self.name = f'effect_of_{risk}_on_{disease_rate}'
        self.configuration_defaults = {self.name: DirectEffect.configuration_defaults['direct_effect']}

    def setup(self, builder):
        relative_risk = builder.configuration[self.name].relative_risk
        self.relative_risk = builder.value.register_value_producer(
            f'{self.name}.relative_risk', source=lambda index: pd.Series(relative_risk, index=index)
        )

        builder.value.register_value_modifier(f'{self.disease_rate}.population_attributable_fraction',
                                              self.population_attributable_fraction)
        builder.value.register_value_modifier(f'{self.disease_rate}',
                                              self.rate_adjustment)
        self.base_risk_exposure = builder.value.get_value(f'{self.risk}.base_proportion_exposed')
        self.actual_risk_exposure = builder.value.get_value(f'{self.risk}.exposure')

    def population_attributable_fraction(self, index):
        exposure = self.base_risk_exposure(index)
        relative_risk = self.relative_risk(index)
        return exposure * (relative_risk - 1) / (exposure * (relative_risk - 1) + 1)

    def rate_adjustment(self, index, rates):
        exposed = self.actual_risk_exposure(index)
        rr = self.relative_risk(index)
        rates[exposed] *= rr[exposed]
        return rates



"""
===========================
Vivarium Command Line Tools
===========================

``vivarium`` provides the tool :command:`simulate` for running simulations
from the command line.  It provides three subcommands:

.. list-table:: ``simulate`` sub-commands
    :header-rows: 1
    :widths: 30, 40

    *   - Name
        - Description
    *   - | **run**
        - | Runs a single simulation from a model specification file.
    *   - | **test**
        - | Runs an example simulation that comes packaged with ``vivarium``.
           | Useful as an installation test.
    *   - | **profile**
        - | Produces a profile of a simulation using the python
          | :mod:`cProfile` module

For more information, see the :ref:`tutorial <cli_tutorial>` on running
simulations from the command line.

.. click:: vivarium.interface.cli:simulate
   :prog: simulate
   :show-nested:

"""
import cProfile
from pathlib import Path
import pstats
from time import time
import shutil

import click
import pandas as pd
from loguru import logger

from vivarium.examples import disease_model
from vivarium.framework.engine import run_simulation
from vivarium.framework.utilities import handle_exceptions
from .utilities import get_output_root, configure_logging_to_file, configure_logging_to_terminal


@click.group()
def simulate():
    """A command line utility for running a single simulation.

    You may initiate a new run with the ``run`` sub-command, initiate a test
    run of a provided model specification with the ``test`` subcommand, or
    profile a simulation run with the ``profile`` subcommand.
    """
    pass


@simulate.command()
@click.argument('model_specification',
                type=click.Path(exists=True, dir_okay=False, resolve_path=True))
@click.option('--results_directory', '-o', type=click.Path(resolve_path=True),
              default=Path('~/vivarium_results/').expanduser(),
              help='The directory to write results to. A folder will be created '
                   'in this directory with the same name as the configuration file.')
@click.option('--verbose', '-v', is_flag=True, help='Report each time step.')
@click.option('--pdb', 'with_debugger', is_flag=True, help='Drop into python debugger if an error occurs.')
def run(model_specification, results_directory, verbose, with_debugger):
    """Run a simulation from the command line.

    The simulation itself is defined by the given MODEL_SPECIFICATION yaml file.

    Within the results directory, which defaults to ~/vivarium_results if none
    is provided, a subdirectory will be created with the same name as the
    MODEL_SPECIFICATION if one does not exist. Results will be written to a
    further subdirectory named after the start time of the simulation run."""
    configure_logging_to_terminal(verbose)

    start = time()
    results_root = get_output_root(results_directory, model_specification)
    results_root.mkdir(parents=True, exist_ok=False)

    configure_logging_to_file(results_root)
    shutil.copy(model_specification, results_root / 'model_specification.yaml')

    main = handle_exceptions(run_simulation, logger, with_debugger)
    override_configuration = {'output_data': {'results_directory': str(results_root)}}
    finished_sim = main(model_specification, configuration=override_configuration)

    idx = pd.Index([finished_sim.configuration.randomness.random_seed], name='random_seed')
    metrics = pd.DataFrame(finished_sim.report(), index=idx)
    metrics['simulation_run_time'] = time() - start
    metrics.to_hdf(results_root / 'output.hdf', key='data')


@simulate.command()
def test():
    """Run a test simulation using the ``disease_model.yaml`` model specification
    provided in the examples directory.
    """
    configure_logging_to_terminal(verbose=True)
    model_specification = disease_model.get_model_specification_path()

    main = handle_exceptions(run_simulation, logger, with_debugger=False)

    main(model_specification)
    click.echo()
    click.secho("Installation test successful!", fg='green')


@simulate.command()
@click.argument('model_specification',
                type=click.Path(exists=True, dir_okay=False, resolve_path=True))
@click.option('--results_directory', '-o', type=click.Path(resolve_path=True),
              default=Path('~/vivarium_results/').expanduser(),
              help='The directory to write results to. A folder will be created '
                   'in this directory with the same name as the configuration file.')
@click.option('--process/--no-process', default=False,
              help=('Automatically process the profile to human readable format  with pstats, '
                    'sorted by cumulative runtime, and dump to a file'))
def profile(model_specification, results_directory, process):
    """Run a simulation based on the provided MODEL_SPECIFICATION and profile
    the run.
    """
    model_specification = Path(model_specification)
    results_directory = Path(results_directory)

    out_stats_file = results_directory / f'{model_specification.name}'.replace('yaml', 'stats')
    command = f'run_simulation("{model_specification}")'
    cProfile.runctx(command, globals=globals(), locals=locals(), filename=out_stats_file)

    if process:
        out_txt_file = results_directory / (out_stats_file.name + '.txt')
        with out_txt_file.open('w') as f:
            p = pstats.Stats(str(out_stats_file), stream=f)
            p.sort_stats('cumulative')
            p.print_stats()




from vivarium.config_tree import ConfigTree

TEST_YAML_ONE = '''
test_section:
    test_key: test_value
    test_key2: test_value2
test_section2:
    test_key: test_value3
    test_key2: test_value4
'''


def test_load_yaml_string():
    d = ConfigTree()
    d.update(TEST_YAML_ONE, source='inline_test')

    assert d.test_section.test_key == 'test_value'
    assert d.test_section.test_key2 == 'test_value2'
    assert d.test_section2.test_key == 'test_value3'


def test_load_yaml_file(tmpdir):
    tmp_file = tmpdir.join('test_file.yaml')
    tmp_file.write(TEST_YAML_ONE)

    d = ConfigTree()
    d.update(str(tmp_file))

    assert d.test_section.test_key == 'test_value'
    assert d.test_section.test_key2 == 'test_value2'
    assert d.test_section2.test_key == 'test_value3'



import numpy as np
import pandas as pd


class Location:

    configuration_defaults = {
        'location': {
            'width': 1000,   # Width of our field
            'height': 1000,  # Height of our field
        }
    }

    def __init__(self):
        self.name = 'location'

    def setup(self, builder):
        self.width = builder.configuration.location.width
        self.height = builder.configuration.location.height

        columns_created = ['x', 'vx', 'y', 'vy']
        builder.population.initializes_simulants(self.on_create_simulants, columns_created)
        self.population_view = builder.population.get_view(columns_created)

    def on_create_simulants(self, pop_data):
        count = len(pop_data.index)
        # Start clustered in the center with small random velocities
        new_population = pd.DataFrame({
            'x': self.width * (0.4 + 0.2 * np.random.random(count)),
            'y': self.height * (0.4 + 0.2 * np.random.random(count)),
            'vx': -0.5 + np.random.random(count),
            'vy': -0.5 + np.random.random(count),
        }, index= pop_data.index)
        self.population_view.update(new_population)



"""
=============
State Machine
=============

A state machine implementation for use in ``vivarium`` simulations.

"""
from enum import Enum

import pandas as pd
import numpy as np


def _next_state(index, event_time, transition_set, population_view):
    """Moves a population between different states using information from a `TransitionSet`.

    Parameters
    ----------
    index : iterable of ints
        An iterable of integer labels for the simulants.
    event_time : pandas.Timestamp
        When this transition is occurring.
    transition_set : TransitionSet
        A set of potential transitions available to the simulants.
    population_view : vivarium.framework.population.PopulationView
        A view of the internal state of the simulation.
    """

    if len(transition_set) == 0 or index.empty:
        return

    outputs, decisions = transition_set.choose_new_state(index)
    groups = _groupby_new_state(index, outputs, decisions)

    if groups:
        for output, affected_index in sorted(groups, key=lambda x: str(x[0])):
            if output == 'null_transition':
                pass
            elif isinstance(output, Transient):
                if not isinstance(output, State):
                    raise ValueError('Invalid transition output: {}'.format(output))
                output.transition_effect(affected_index, event_time, population_view)
                output.next_state(affected_index, event_time, population_view)
            elif isinstance(output, State):
                output.transition_effect(affected_index, event_time, population_view)
            else:
                raise ValueError('Invalid transition output: {}'.format(output))


def _groupby_new_state(index, outputs, decisions):
    """Groups the simulants in the index by their new output state.

    Parameters
    ----------
    index : iterable of ints
        An iterable of integer labels for the simulants.
    outputs : iterable
        A list of possible output states.
    decisions : `pandas.Series`
        A series containing the name of the next state for each simulant in the index.

    Returns
    -------
    iterable of 2-tuples
        The first item in each tuple is the name of an output state and the second item
        is a `pandas.Index` representing the simulants to transition into that state.
    """
    output_map = {o: i for i, o in enumerate(outputs)}
    groups = pd.Series(index).groupby([output_map[d] for d in decisions])
    results = [(outputs[i], pd.Index(sub_group.values)) for i, sub_group in groups]
    selected_outputs = [o for o, _ in results]
    for output in outputs:
        if output not in selected_outputs:
            results.append((output, pd.Index([])))
    return results


class Trigger(Enum):
    NOT_TRIGGERED = 0
    START_INACTIVE = 1
    START_ACTIVE = 2


def _process_trigger(trigger):
    if trigger == Trigger.NOT_TRIGGERED:
        return None, False
    elif trigger == Trigger.START_INACTIVE:
        return pd.Index([]), False
    elif trigger == Trigger.START_ACTIVE:
        return pd.Index([]), True
    else:
        raise ValueError("Invalid trigger state provided: {}".format(trigger))


class Transition:
    """A process by which an entity might change into a particular state.

    Parameters
    ----------
    input_state: State
        The start state of the entity that undergoes the transition.
    output_state : State
        The end state of the entity that undergoes the transition.
    probability_func : Callable
        A method or function that describing the probability of this transition occurring.
    """
    def __init__(self, input_state, output_state, probability_func=lambda index: pd.Series(1, index=index),
                 triggered=Trigger.NOT_TRIGGERED):
        self.input_state = input_state
        self.output_state = output_state
        self._probability = probability_func
        self._active_index, self.start_active = _process_trigger(triggered)

    @property
    def name(self):
        return f"transition.{self.input_state.name}.{self.output_state.name}"

    def setup(self, builder):
        pass

    def set_active(self, index):
        if self._active_index is None:
            raise ValueError("This transition is not triggered.  An active index cannot be set or modified.")
        else:
            self._active_index = self._active_index.union(pd.Index(index))

    def set_inactive(self, index):
        if self._active_index is None:
            raise ValueError("This transition is not triggered.  An active index cannot be set or modified.")
        else:
            self._active_index = self._active_index.difference(pd.Index(index))

    def probability(self, index):
        if self._active_index is None:
            return self._probability(index)

        index = pd.Index(index)
        activated_index = self._active_index.intersection(index)
        null_index = index.difference(self._active_index)
        activated = pd.Series(self._probability(activated_index), index=activated_index)
        null = pd.Series(np.zeros(len(null_index), dtype=float), index=null_index)
        return activated.append(null)

    def __repr__(self):
        return f'Transition({self.input_state}, {self.output_state})'


class State:
    """An abstract representation of a particular position in a finite and discrete state space.

    Attributes
    ----------
    state_id : str
        The name of this state. This should be unique
    transition_set : `TransitionSet`
        A container for potential transitions out of this state.
    """
    def __init__(self, state_id):
        self.state_id = state_id
        self.transition_set = TransitionSet(self.name)
        self._model = None
        self._sub_components = [self.transition_set]

    @property
    def name(self):
        return f"state.{self.state_id}"

    @property
    def sub_components(self):
        return self._sub_components

    def setup(self, builder):
        pass

    def next_state(self, index, event_time, population_view):
        """Moves a population between different states using information this state's `transition_set`.

        Parameters
        ----------
        index : iterable of ints
            An iterable of integer labels for the simulants.
        event_time : pandas.Timestamp
            When this transition is occurring.
        population_view : vivarium.framework.population.PopulationView
            A view of the internal state of the simulation.
        """
        return _next_state(index, event_time, self.transition_set, population_view)

    def transition_effect(self, index, event_time, population_view):
        """Updates the simulation state and triggers any side-effects associated with entering this state.

        Parameters
        ----------
        index : iterable of ints
            An iterable of integer labels for the simulants.
        event_time : pandas.Timestamp
            The time at which this transition occurs.
        population_view : `vivarium.framework.population.PopulationView`
            A view of the internal state of the simulation.
        """
        population_view.update(pd.Series(self.state_id, index=index))
        self._transition_side_effect(index, event_time)

    def cleanup_effect(self, index, event_time):
        self._cleanup_effect(index, event_time)

    def add_transition(self, output,
                       probability_func=lambda index: np.ones(len(index), dtype=float),
                       triggered=Trigger.NOT_TRIGGERED):
        """Builds a transition from this state to the given state.

        output : State
            The end state after the transition.
        """
        t = Transition(self, output, probability_func=probability_func, triggered=triggered)
        self.transition_set.append(t)
        return t

    def allow_self_transitions(self):
        self.transition_set.allow_null_transition = True

    def _transition_side_effect(self, index, event_time):
        pass

    def _cleanup_effect(self, index, event_time):
        pass

    def __repr__(self):
        return f'State({self.state_id})'


class Transient:
    """Used to tell _next_state to transition a second time."""
    pass


class TransientState(State, Transient):
    def __repr__(self):
        return f'TransientState({self.state_id})'


class TransitionSet:
    """A container for state machine transitions.

    Parameters
    ----------
    state_name: object
        The unique name of the state that instantiated this TransitionSet. Typically
        a string but any object implementing __str__ will do.
    iterable : iterable
        Any iterable whose elements are `Transition` objects.
    allow_null_transition : bool, optional
    """
    def __init__(self, state_name, *transitions, allow_null_transition=False):
        self._state_name = state_name
        self.allow_null_transition = allow_null_transition
        self.transitions = []
        self._sub_components = self.transitions

        self.extend(transitions)

    @property
    def name(self):
        return f'transition_set.{self._state_name}'

    @property
    def sub_components(self):
        return self._sub_components

    def setup(self, builder):
        """Performs this component's simulation setup and return sub-components.

        Parameters
        ----------
        builder : `engine.Builder`
            Interface to several simulation tools including access to common random
            number generation, in particular.

        Returns
        -------
        iterable
            This component's sub-components.
        """
        self.random = builder.randomness.get_stream(self.name)

    def choose_new_state(self, index):
        """Chooses a new state for each simulant in the index.

        Parameters
        ----------
        index : iterable of ints
            An iterable of integer labels for the simulants.

        Returns
        -------
        outputs : list
            The possible end states of this set of transitions.
        decisions: `pandas.Series`
            A series containing the name of the next state for each simulant in the index.
        """
        outputs, probabilities = zip(*[(transition.output_state, np.array(transition.probability(index)))
                                       for transition in self.transitions])
        probabilities = np.transpose(probabilities)
        outputs, probabilities = self._normalize_probabilities(outputs, probabilities)
        return outputs, self.random.choice(index, outputs, probabilities)

    def _normalize_probabilities(self, outputs, probabilities):
        """Normalize probabilities to sum to 1 and add a null transition if desired.

        Parameters
        ----------
        outputs : iterable
            List of possible end states corresponding to this containers transitions.
        probabilities : iterable of iterables
            A set of probability weights whose columns correspond to the end states in `outputs`
            and whose rows correspond to each simulant undergoing the transition.

        Returns
        -------
        outputs: list
            The original output list expanded to include a null transition (a transition back
            to the starting state) if requested.
        probabilities : ndarray
            The original probabilities rescaled to sum to 1 and potentially expanded to
            include a null transition weight.
        """
        outputs = list(outputs)

        # This is mainly for flexibility with the triggered transitions.
        # We may have multiple out transitions from a state where one of them
        # is gated until some criteria is met.  After the criteria is
        # met, the gated transition becomes the default (likely as opposed
        # to a self transition).
        default_transition_count = np.sum(probabilities == 1, axis=1)
        if np.any(default_transition_count > 1):
            raise ValueError("Multiple transitions specified with probability 1.")
        has_default = default_transition_count == 1
        total = np.sum(probabilities, axis=1)
        probabilities[has_default] /= total[has_default, np.newaxis]

        total = np.sum(probabilities, axis=1)  # All totals should be ~<= 1 at this point.
        if self.allow_null_transition:
            if np.any(total > 1+1e-08):  # Accommodate rounding errors
                raise ValueError(f"Null transition requested with un-normalized "
                                 f"probability weights: {probabilities}")
            total[total > 1] = 1  # Correct allowed rounding errors.
            probabilities = np.concatenate([probabilities, (1-total)[:, np.newaxis]], axis=1)
            outputs.append('null_transition')
        else:
            if np.any(total == 0):
                raise ValueError("No valid transitions for some simulants.")
            else:  # total might be less than zero in some places
                probabilities /= total[:, np.newaxis]

        return outputs, probabilities

    def append(self, transition):
        if not isinstance(transition, Transition):
            raise TypeError(
                'TransitionSet must contain only Transition objects. Check constructor arguments: {}'.format(self))
        self.transitions.append(transition)

    def extend(self, transitions):
        for transition in transitions:
            self.append(transition)

    def __iter__(self):
        return iter(self.transitions)

    def __len__(self):
        return len(self.transitions)

    def __repr__(self):
        return f"TransitionSet(transitions={[x for x in self.transitions]})"

    def __hash__(self):
        return hash(id(self))


class Machine:
    """A collection of states and transitions between those states.

    Attributes
    ----------
    states : iterable of `State` objects
        The collection of states represented by this state machine.
    state_column : str
        A label for the piece of simulation state governed by this state machine.
    population_view : `pandas.DataFrame`
        A view of the internal state of the simulation.
    """
    def __init__(self, state_column, states=None):
        self.states = []
        self.state_column = state_column
        if states:
            self.add_states(states)

    @property
    def name(self):
        return f"machine.{self.state_column}"

    @property
    def sub_components(self):
        return self.states

    def setup(self, builder):
        """Performs this component's simulation setup and return sub-components.

        Parameters
        ----------
        builder : `engine.Builder`
            Interface to several simulation tools including access to common random
            number generation, in particular.

        Returns
        -------
        iterable
            This component's sub-components.
        """
        self.population_view = builder.population.get_view([self.state_column])

    def add_states(self, states):
        for state in states:
            self.states.append(state)
            state._model = self.state_column

    def transition(self, index, event_time):
        """Finds the population in each state and moves them to the next state.

        Parameters
        ----------
        index : iterable of ints
            An iterable of integer labels for the simulants.
        event_time : pandas.Timestamp
            The time at which this transition occurs.
        """
        for state, affected in self._get_state_pops(index):
            if not affected.empty:
                state.next_state(affected.index, event_time, self.population_view.subview([self.state_column]))

    def cleanup(self, index, event_time):
        for state, affected in self._get_state_pops(index):
            if not affected.empty:
                state.cleanup_effect(affected.index, event_time)

    def to_dot(self):
        """Produces a ball and stick graph of this state machine.

        Returns
        -------
        `graphviz.Digraph`
            A ball and stick visualization of this state machine.
        """
        from graphviz import Digraph
        dot = Digraph(format='png')
        for state in self.states:
            if isinstance(state, TransientState):
                dot.node(state.state_id, style='dashed')
            else:
                dot.node(state.state_id)
            for transition in state.transition_set:
                dot.edge(state.state_id, transition.output_state.state_id, transition.name)
        return dot

    def _get_state_pops(self, index):
        population = self.population_view.get(index)
        return [[state, population[population[self.state_column] == state.state_id]] for state in self.states]

    def __repr__(self):
        return f"Machine(state_column= {self.state_column})"



import pytest

import pandas as pd
import numpy as np
import itertools

from vivarium.interpolation import Interpolation, validate_parameters, check_data_complete, Order0Interp


def make_bin_edges(data: pd.DataFrame, col: str) -> pd.DataFrame:
    """ Given a dataframe and a column containing midpoints, construct
    equally sized bins around midpoints.
    """
    mid_pts = data[[col]].drop_duplicates().sort_values(by=col).reset_index(drop=True)
    mid_pts['shift'] = mid_pts[col].shift()

    mid_pts['left'] = mid_pts.apply(lambda row: (row[col] if pd.isna(row['shift'])
                                                 else 0.5 * (row[col] + row['shift'])), axis=1)

    mid_pts['right'] = mid_pts['left'].shift(-1)
    mid_pts['right'] = mid_pts.right.fillna(mid_pts.right.max() + mid_pts.left.tolist()[-1] - mid_pts.left.tolist()[-2])

    data = data.copy()
    idx = data.index

    data = data.set_index(col, drop=False)
    mid_pts = mid_pts.set_index(col, drop=False)

    data[[col, f'{col}_left', f'{col}_right']] = mid_pts[[col, 'left', 'right']]

    return data.set_index(idx)


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_1d_interpolation():
    df = pd.DataFrame({'a': np.arange(100), 'b': np.arange(100), 'c': np.arange(100, 0, -1)})
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, (), ('a',), 1, True)

    query = pd.DataFrame({'a': np.arange(100, step=0.01)})

    assert np.allclose(query.a, i(query).b)
    assert np.allclose(100-query.a, i(query).c)


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_age_year_interpolation():
    years = list(range(1990, 2010))
    ages = list(range(0, 90))
    pops = np.array(ages)*11.1
    data = []
    for age, pop in zip(ages, pops):
        for year in years:
            for sex in ['Male', 'Female']:
                data.append({'age': age, 'sex': sex, 'year': year, 'pop': pop})
    df = pd.DataFrame(data)

    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, ('sex', 'age'), ('year',), 1, True)
    query = pd.DataFrame({'year': [1990, 1990], 'age': [35, 35], 'sex': ['Male', 'Female']})
    assert np.allclose(i(query), 388.5)


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_interpolation_called_missing_key_col():
    a = [range(1990, 1995), range(25, 30), ['Male', 'Female']]
    df = pd.DataFrame(list(itertools.product(*a)), columns=['year', 'age', 'sex'])
    df['pop'] = df.age * 11.1
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input
    i = Interpolation(df, ['sex',], ['year','age'], 1, True)
    query = pd.DataFrame({'year': [1990, 1990], 'age': [35, 35]})
    with pytest.raises(ValueError):
        i(query)


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_interpolation_called_missing_param_col():
    a = [range(1990, 1995), range(25, 30), ['Male', 'Female']]
    df = pd.DataFrame(list(itertools.product(*a)), columns=['year', 'age', 'sex'])
    df['pop'] = df.age * 11.1
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input
    i = Interpolation(df, ['sex',], ['year','age'], 1, True)
    query = pd.DataFrame({'year': [1990, 1990], 'sex': ['Male', 'Female']})
    with pytest.raises(ValueError):
        i(query)


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_2d_interpolation():
    a = np.mgrid[0:5, 0:5][0].reshape(25)
    b = np.mgrid[0:5, 0:5][1].reshape(25)
    df = pd.DataFrame({'a': a, 'b': b, 'c': b, 'd': a})
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, (), ('a', 'b'), 1, True)

    query = pd.DataFrame({'a': np.arange(4, step=0.01), 'b': np.arange(4, step=0.01)})

    assert np.allclose(query.b, i(query).c)
    assert np.allclose(query.a, i(query).d)


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_interpolation_with_categorical_parameters():
    a = ['one']*100 + ['two']*100
    b = np.append(np.arange(100), np.arange(100))
    c = np.append(np.arange(100), np.arange(100, 0, -1))
    df = pd.DataFrame({'a': a, 'b': b, 'c': c})
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, ('a',), ('b',), 1, True)

    query_one = pd.DataFrame({'a': 'one', 'b': np.arange(100, step=0.01)})
    query_two = pd.DataFrame({'a': 'two', 'b': np.arange(100, step=0.01)})

    assert np.allclose(np.arange(100, step=0.01), i(query_one).c)

    assert np.allclose(np.arange(100, 0, step=-0.01), i(query_two).c)


def test_order_zero_2d():
    a = np.mgrid[0:5, 0:5][0].reshape(25)
    b = np.mgrid[0:5, 0:5][1].reshape(25)
    df = pd.DataFrame({'a': a + 0.5, 'b': b + 0.5, 'c': b*3, 'garbage': ['test']*len(a)})
    df = make_bin_edges(df, 'a')
    df = make_bin_edges(df, 'b')
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, ('garbage',), [('a', 'a_left', 'a_right'), ('b', 'b_left', 'b_right')],
                      order=0, extrapolate=True)

    column = np.arange(0.5, 4, step=0.011)
    query = pd.DataFrame({'a': column, 'b': column, 'garbage': ['test']*(len(column))})

    assert np.allclose(query.b.astype(int) * 3, i(query).c)


def test_order_zero_2d_fails_on_extrapolation():
    a = np.mgrid[0:5, 0:5][0].reshape(25)
    b = np.mgrid[0:5, 0:5][1].reshape(25)
    df = pd.DataFrame({'a': a + 0.5, 'b': b + 0.5, 'c': b*3, 'garbage': ['test']*len(a)})
    df = make_bin_edges(df, 'a')
    df = make_bin_edges(df, 'b')
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, ('garbage',), [('a', 'a_left', 'a_right'), ('b', 'b_left', 'b_right')],
                      order=0, extrapolate=False)

    column = np.arange(4, step=0.011)
    query = pd.DataFrame({'a': column, 'b': column, 'garbage': ['test']*(len(column))})

    with pytest.raises(ValueError) as error:
        i(query)

    message = error.value.args[0]

    assert 'Extrapolation' in message and 'a' in message


def test_order_zero_1d_no_extrapolation():
    s = pd.Series({0: 0, 1: 1}).reset_index()
    s = make_bin_edges(s, 'index')
    f = Interpolation(s, tuple(), [['index', 'index_left', 'index_right']], order=0, extrapolate=False)

    assert f(pd.DataFrame({'index': [0]}))[0][0] == 0, 'should be precise at index values'
    assert f(pd.DataFrame({'index': [0.999]}))[0][0] == 1

    with pytest.raises(ValueError) as error:
        f(pd.DataFrame({'index': [1]}))

    message = error.value.args[0]
    assert 'Extrapolation' in message and 'index' in message


def test_order_zero_1d_constant_extrapolation():
    s = pd.Series({0: 0, 1: 1}).reset_index()
    s = make_bin_edges(s, 'index')
    f = Interpolation(s, tuple(), [['index', 'index_left', 'index_right']], order=0, extrapolate=True)

    assert f(pd.DataFrame({'index': [1]}))[0][0] == 1
    assert f(pd.DataFrame({'index': [2]}))[0][0] == 1, 'should be constant extrapolation outside of input range'
    assert f(pd.DataFrame({'index': [-1]}))[0][0] == 0


def test_validate_parameters__empty_data():
    with pytest.raises(ValueError) as error:
        validate_parameters(pd.DataFrame(columns=["age_left", "age_right",
                                                  "sex", "year_left", "year_right", "value"]), ["sex"],
                            [("age", "age_left", "age_right"),
                             ["year", "year_left", "year_right"]])
    message = error.value.args[0]
    assert 'empty' in message


def test_check_data_complete_gaps():
    data = pd.DataFrame({'year_start': [1990, 1990, 1995, 1995],
                         'year_end': [1995, 1995, 2000, 2000],
                         'age_start': [16, 10, 10, 16],
                         'age_end': [20, 15, 15, 20],})

    with pytest.raises(NotImplementedError) as error:
        check_data_complete(data, [('year', 'year_start', 'year_end'), ['age', 'age_start', 'age_end']])

    message = error.value.args[0]

    assert "age_start" in message and "age_end" in message


def test_check_data_complete_overlap():
    data = pd.DataFrame({'year_start': [1995, 1995, 2000, 2005, 2010],
                         'year_end': [2000, 2000, 2005, 2010, 2015]})

    with pytest.raises(ValueError) as error:
        check_data_complete(data, [('year', 'year_start', 'year_end')])

    message = error.value.args[0]

    assert "year_start" in message and "year_end" in message


def test_check_data_missing_combos():
    data = pd.DataFrame({'year_start': [1990, 1990, 1995],
                         'year_end': [1995, 1995, 2000],
                         'age_start': [10, 15, 10],
                         'age_end': [15, 20, 15]})

    with pytest.raises(ValueError) as error:
        check_data_complete(data, [['year', 'year_start', 'year_end'], ('age', 'age_start', 'age_end')])

    message = error.value.args[0]

    assert 'combination' in message


def test_order0interp():
    data = pd.DataFrame({'year_start': [1990, 1990, 1990, 1990, 1995, 1995, 1995, 1995],
                         'year_end': [1995, 1995, 1995, 1995, 2000, 2000, 2000, 2000],
                         'age_start': [15, 10, 10, 15, 10, 10, 15, 15],
                         'age_end': [20, 15, 15, 20, 15, 15, 20, 20],
                         'height_start': [140, 160, 140, 160, 140, 160, 140, 160],
                         'height_end': [160, 180, 160, 180, 160, 180, 160, 180],
                         'value': [5, 3, 1, 7, 8, 6, 4, 2]})

    interp = Order0Interp(data, [('age', 'age_start', 'age_end'),
                                 ('year', 'year_start', 'year_end'),
                                 ('height', 'height_start', 'height_end'),]
                          , ['value'], True)

    interpolants = pd.DataFrame({'age': [12, 17, 8, 24, 12],
                                 'year': [1992, 1998, 1985, 1992, 1992],
                                 'height': [160, 145, 140, 179, 160]})

    result = interp(interpolants)
    assert result.equals(pd.DataFrame({'value': [3, 4, 1, 7, 3]}))


def test_order_zero_1d_with_key_column():
    data = pd.DataFrame({'year_start': [1990, 1990, 1995, 1995],
                         'year_end': [1995, 1995, 2000, 2000],
                         'sex': ['Male', 'Female', 'Male', 'Female'],
                         'value_1': [10, 7, 2, 12],
                         'value_2': [1200, 1350, 1476, 1046]})

    i = Interpolation(data, ['sex',], [('year', 'year_start', 'year_end'),], 0, True)

    query = pd.DataFrame({'year': [1992, 1993,],
                          'sex': ['Male', 'Female']})

    expected_result = pd.DataFrame({'value_1': [10.0, 7.0],
                                    'value_2': [1200.0, 1350.0]})

    assert i(query).equals(expected_result)


def test_order_zero_non_numeric_values():
    data = pd.DataFrame({'year_start': [1990, 1990],
                         'year_end': [1995, 1995],
                         'age_start': [15, 24,],
                         'age_end': [24, 30],
                         'value_1': ['blue', 'red']})

    i = Interpolation(data, tuple(), [('year', 'year_start', 'year_end'), ('age', 'age_start', 'age_end')], 0, True)

    query = pd.DataFrame({'year': [1990, 1990],
                          'age': [15, 24,]},
                         index=[1, 0])

    expected_result = pd.DataFrame({'value_1': ['blue', 'red']},
                                   index=[1, 0])

    assert i(query).equals(expected_result)


def test_order_zero_3d_with_key_col():
    data = pd.DataFrame({'year_start': [1990, 1990, 1990, 1990, 1995, 1995, 1995, 1995]*2,
                         'year_end': [1995, 1995, 1995, 1995, 2000, 2000, 2000, 2000]*2,
                         'age_start': [15, 10, 10, 15, 10, 10, 15, 15]*2,
                         'age_end': [20, 15, 15, 20, 15, 15, 20, 20]*2,
                         'height_start': [140, 160, 140, 160, 140, 160, 140, 160]*2,
                         'height_end': [160, 180, 160, 180, 160, 180, 160, 180]*2,
                         'sex': ['Male']*8+['Female']*8,
                         'value': [5, 3, 1, 7, 8, 6, 4, 2, 6, 4, 2, 8, 9, 7, 5, 3]})

    interp = Interpolation(data, ('sex',),
                           [('age', 'age_start', 'age_end'),
                            ('year', 'year_start', 'year_end'),
                            ('height', 'height_start', 'height_end')], 0, True)

    interpolants = pd.DataFrame({'age': [12, 17, 8, 24, 12],
                                 'year': [1992, 1998, 1985, 1992, 1992],
                                 'height': [160, 145, 140, 185, 160],
                                 'sex': ['Male', 'Female', 'Female', 'Male', 'Male']},
                                index=[10, 4, 7, 0, 9])

    result = interp(interpolants)
    assert result.equals(pd.DataFrame({'value': [3.0, 5.0, 2.0, 7.0, 3.0]}, index=[10, 4, 7, 0, 9]))


def test_order_zero_diff_bin_sizes():
    data = pd.DataFrame({'year_start': [1990, 1995, 1996, 2005, 2005.5,],
                         'year_end': [1995, 1996, 2005, 2005.5, 2010],
                         'value': [1, 5, 2.3, 6, 100]})

    i = Interpolation(data, tuple(), [('year', 'year_start', 'year_end')], 0, False)

    query = pd.DataFrame({'year': [2007, 1990, 2005.4, 1994, 2004, 1995, 2002, 1995.5, 1996]})

    expected_result = pd.DataFrame({'value': [100, 1, 6, 1, 2.3, 5, 2.3, 5, 2.3]})

    assert i(query).equals(expected_result)


def test_order_zero_given_call_column():
    data = pd.DataFrame({'year_start': [1990, 1995, 1996, 2005, 2005.5,],
                         'year_end': [1995, 1996, 2005, 2005.5, 2010],
                         'year': [1992.5, 1995.5, 2000, 2005.25, 2007.75],
                         'value': [1, 5, 2.3, 6, 100]})

    i = Interpolation(data, tuple(), [('year', 'year_start', 'year_end')], 0, False)

    query = pd.DataFrame({'year': [2007, 1990, 2005.4, 1994, 2004, 1995, 2002, 1995.5, 1996]})

    expected_result = pd.DataFrame({'value': [100, 1, 6, 1, 2.3, 5, 2.3, 5, 2.3]})

    assert i(query).equals(expected_result)


import numpy as np
import pandas as pd

from vivarium import InteractiveContext
from vivarium.testing_utilities import build_table, TestPopulation
from vivarium.framework.lookup import validate_parameters
import pytest


@pytest.mark.skip(reason='only order 0 interpolation with age bin edges currently supported')
def test_interpolated_tables(base_config):
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years = build_table(lambda age, sex, year: year, year_start, year_end)
    ages = build_table(lambda age, sex, year: age, year_start, year_end)
    one_d_age = ages.copy()
    del one_d_age['year']
    one_d_age = one_d_age.drop_duplicates()
    base_config.update({'population': {'population_size': 10000},
                        'interpolation': {'order': 1}})  # the results we're checking later assume interp order 1

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    years = manager.build_table(years, key_columns=('sex',), parameter_columns=('age', 'year',), value_columns=None)
    ages = manager.build_table(ages, key_columns=('sex',), parameter_columns=('age', 'year',), value_columns=None)
    one_d_age = manager.build_table(one_d_age, key_columns=('sex',), parameter_columns=('age',), value_columns=None)

    pop = simulation.get_population(untracked=True)
    result_years = years(pop.index)
    result_ages = ages(pop.index)
    result_ages_1d = one_d_age(pop.index)

    fractional_year = simulation._clock.time.year
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)
    assert np.allclose(result_ages, pop.age)
    assert np.allclose(result_ages_1d, pop.age)

    simulation._clock._time += pd.Timedelta(30.5 * 125, unit='D')
    simulation._population._population.age += 125/12

    result_years = years(pop.index)
    result_ages = ages(pop.index)
    result_ages_1d = one_d_age(pop.index)

    fractional_year = simulation._clock.time.year
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)
    assert np.allclose(result_ages, pop.age)
    assert np.allclose(result_ages_1d, pop.age)


@pytest.mark.skip(reason='only order 0 interpolation with age bin edges currently supported')
def test_interpolated_tables_without_uninterpolated_columns(base_config):
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years = build_table(lambda age, sex, year: year, year_start, year_end)
    del years['sex']
    years = years.drop_duplicates()
    base_config.update({'population': {'population_size': 10000},
                        'interpolation': {'order': 1}})  # the results we're checking later assume interp order 1

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    years = manager.build_table(years, key_columns=(), parameter_columns=('year', 'age',), value_columns=None)

    result_years = years(simulation.get_population().index)

    fractional_year = simulation._clock.time.year
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)

    simulation._clock._time += pd.Timedelta(30.5 * 125, unit='D')

    result_years = years(simulation.get_population().index)

    fractional_year = simulation._clock.time.year
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)


def test_interpolated_tables__exact_values_at_input_points(base_config):
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years = build_table(lambda age, sex, year: year, year_start, year_end)
    input_years = years.year_start.unique()
    base_config.update({'population': {'population_size': 10000}})

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    years = manager._build_table(years, key_columns=['sex'], parameter_columns=['age', 'year'], value_columns=None)

    for year in input_years:
        simulation._clock._time = pd.Timestamp(year, 1, 1)
        assert np.allclose(years(simulation.get_population().index),
                           simulation._clock.time.year + 1/365)


def test_lookup_table_scalar_from_list(base_config):
    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    table = (manager._build_table((1, 2), key_columns=None, parameter_columns=None,
                                  value_columns=['a', 'b'])(simulation.get_population().index))

    assert isinstance(table, pd.DataFrame)
    assert table.columns.values.tolist() == ['a', 'b']
    assert np.all(table.a == 1)
    assert np.all(table.b == 2)


def test_lookup_table_scalar_from_single_value(base_config):
    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    table = (manager._build_table(1, key_columns=None, parameter_columns=None,
                                  value_columns=['a'])(simulation.get_population().index))
    assert isinstance(table, pd.Series)
    assert np.all(table == 1)


def test_invalid_data_type_build_table(base_config):
    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    with pytest.raises(TypeError):
        manager._build_table('break', key_columns=None, parameter_columns=None, value_columns=None)


def test_lookup_table_interpolated_return_types(base_config):
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    data = build_table(lambda age, sex, year: year, year_start, year_end)

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    table = (manager._build_table(data, key_columns=['sex'], parameter_columns=['age', 'year'],
                                  value_columns=None)(simulation.get_population().index))
    # make sure a single value column is returned as a series
    assert isinstance(table, pd.Series)

    # now add a second value column to make sure the result is a df
    data['value2'] = data.value
    table = (manager._build_table(data, key_columns=['sex'],
                                  parameter_columns=['age', 'year'],
                                  value_columns=None)(simulation.get_population().index))

    assert isinstance(table, pd.DataFrame)


@pytest.mark.parametrize('data', [None, pd.DataFrame(), pd.DataFrame(columns=['a', 'b', 'c']), [], tuple()])
def test_validate_parameters_no_data(data):
    with pytest.raises(ValueError, match='supply some data'):
        validate_parameters(data, [], [], [])


@pytest.mark.parametrize('key_cols, param_cols, val_cols, match',
                         [(None, None, None, 'supply value_columns'),
                          (None, None, [], 'supply value_columns'),
                          (None, None, ['a', 'b'], 'match the number of values')])
def test_validate_parameters_error_scalar_data(key_cols, param_cols, val_cols, match):
    with pytest.raises(ValueError, match=match):
        validate_parameters([1, 2, 3], key_cols, param_cols, val_cols)


@pytest.mark.parametrize('key_cols, param_cols, val_cols, match',
                         [(['a', 'b'], ['b'], ['c'], 'no overlap'),
                          ([], ['b'], ['c'], 'do not match')])
def test_validate_parameters_error_dataframe(key_cols, param_cols, val_cols, match):
    data = pd.DataFrame({'a': [1, 2], 'b_start': [0, 5], 'b_end': [5, 10], 'c': [100, 150]})
    with pytest.raises(ValueError, match=match):
        validate_parameters(data, key_cols, param_cols, val_cols)


@pytest.mark.parametrize('data', ['FAIL', pd.Interval(5, 10), '2019-05-17', {'a': 5, 'b': 10}])
def test_validate_parameters_fail_other_data(data):
    with pytest.raises(TypeError, match='only allowable types'):
        validate_parameters(data, [], [], [])


@pytest.mark.parametrize('key_cols, param_cols, val_cols',
                         [(None, None, ['a', 'b', 'c']),
                          (None, ['d'], ['one', 'two', 'three']),
                          (['KEY'], None, ['a', 'b', 'c']),
                          (['KEY'], ['d'], ['a', 'b', 'c'])])
def test_validate_parameters_pass_scalar_data(key_cols, param_cols, val_cols):
    validate_parameters([1, 2, 3], key_cols, param_cols, val_cols)


@pytest.mark.parametrize('key_cols, param_cols, val_cols',
                         [(['a'], ['b'], ['c']),
                          ([], ['b'], ['c', 'a']),
                          ([], ['b'], ['a', 'c'])])
def test_validate_parameters_pass_dataframe(key_cols, param_cols, val_cols):
    data = pd.DataFrame({'a': [1, 2], 'b_start': [0, 5], 'b_end': [5, 10], 'c': [100, 150]})
    validate_parameters(data, key_cols, param_cols, val_cols)



import pytest

from vivarium.framework.plugins import PluginManager, PluginConfigurationError, DEFAULT_PLUGINS
from vivarium.framework.time import DateTimeClock, TimeInterface
from vivarium.framework.components import ComponentConfigurationParser

from .components.mocks import MockComponentA

plugin_config = {'george': {'controller': 'big_brother',
                            'builder_interface': 'minipax'}}


@pytest.fixture
def test_plugin_manager(model_specification):
    model_specification.plugins.optional.update(plugin_config)
    return PluginManager(model_specification.plugins)


def test_PluginManager_initializaiton(model_specification):
    model_specification.plugins.optional.update(plugin_config)
    plugin_manager = PluginManager(model_specification.plugins)

    assert model_specification.plugins.to_dict() == plugin_manager._plugin_configuration.to_dict()
    assert not plugin_manager._plugins


def test_PluginManager__lookup_fail(test_plugin_manager):
    with pytest.raises(PluginConfigurationError):
        test_plugin_manager._lookup('bananas')


def test_PluginManager__lookup(test_plugin_manager):
    for plugin in ['component_manager', 'clock', 'component_configuration_parser']:
        assert test_plugin_manager._lookup(plugin).to_dict() == DEFAULT_PLUGINS['plugins']['required'][plugin]

    assert test_plugin_manager._lookup('george').to_dict() == plugin_config['george']


def test_PluginManager__get_fail(test_plugin_manager, mocker):
    import_by_path_mock = mocker.patch('vivarium.framework.plugins.import_by_path')

    def err(path):
        if path == 'vivarium.framework.time.DateTimeClock':
            raise ValueError()

    import_by_path_mock.side_effect = err

    with pytest.raises(PluginConfigurationError):
        test_plugin_manager._get('clock')

    def err(path):
        if path == 'vivarium.framework.time.DateTimeClock':
            return lambda: 'fake_controller'
        elif path == 'vivarium.framework.time.TimeInterface':
            raise ValueError()

    import_by_path_mock.side_effect = err

    with pytest.raises(PluginConfigurationError):
        test_plugin_manager._get('clock')


def test_PluginManager__get(test_plugin_manager):
    time_components = test_plugin_manager._get('clock')

    assert isinstance(time_components['controller'], DateTimeClock)
    assert isinstance(time_components['builder_interface'], TimeInterface)

    parser_components = test_plugin_manager._get('component_configuration_parser')

    assert isinstance(parser_components['controller'], ComponentConfigurationParser)
    assert parser_components['builder_interface'] is None


def test_PluginManager_get_plugin(test_plugin_manager):
    assert test_plugin_manager._plugins == {}
    clock = test_plugin_manager.get_plugin('clock')
    assert isinstance(clock, DateTimeClock)
    assert test_plugin_manager._plugins['clock']['controller'] is clock


def test_PluginManager_get_plugin_interface(test_plugin_manager):
    assert test_plugin_manager._plugins == {}
    clock_interface = test_plugin_manager.get_plugin_interface('clock')
    assert isinstance(clock_interface, TimeInterface)
    assert test_plugin_manager._plugins['clock']['builder_interface'] is clock_interface


def test_PluginManager_get_optional_controllers(test_plugin_manager, mocker):
    import_by_path_mock = mocker.patch('vivarium.framework.plugins.import_by_path')
    component = MockComponentA('george')

    def import_by_path_side_effect(arg):
        if arg == 'big_brother':
            return lambda: component
        else:
            return lambda _: component

    import_by_path_mock.side_effect = import_by_path_side_effect
    assert test_plugin_manager.get_optional_controllers() == {'george': component}
    assert import_by_path_mock.mock_calls == [mocker.call(plugin_config['george']['controller']),
                                              mocker.call(plugin_config['george']['builder_interface'])]


def test_PluginManager_get_optional_interfaces(test_plugin_manager, mocker):
    import_by_path_mock = mocker.patch('vivarium.framework.plugins.import_by_path')
    component = MockComponentA('george')

    def import_by_path_side_effect(arg):
        if arg == 'big_brother':
            return lambda: component
        else:
            return lambda _: component

    import_by_path_mock.side_effect = import_by_path_side_effect
    assert test_plugin_manager.get_optional_interfaces() == {'george': component}
    assert import_by_path_mock.mock_calls == [mocker.call(plugin_config['george']['controller']),
                                              mocker.call(plugin_config['george']['builder_interface'])]



import numpy as np
import pandas as pd


class Population:

    configuration_defaults = {
        'population': {
            'colors': ['red', 'blue'],
        }
    }

    def __init__(self):
        self.name = 'population'

    def setup(self, builder):
        self.colors = builder.configuration.population.colors

        columns_created = ['color', 'entrance_time']
        builder.population.initializes_simulants(self.on_initialize_simulants, columns_created)
        self.population_view = builder.population.get_view(columns_created)

    def on_initialize_simulants(self, pop_data):
        new_population = pd.DataFrame({
            'color': np.random.choice(self.colors, len(pop_data.index)),
            'entrance_time': pop_data.creation_time,
        }, index=pop_data.index)
        self.population_view.update(new_population)
